import pytest
from gpaw.mpi import world
from gpaw.utilities import compiled_with_sl
import numpy as np
from ase import Atoms
from ase.build import bulk
from ase.dft.kpoints import monkhorst_pack

from gpaw import GPAW, FermiDirac
from gpaw.mpi import serial_comm
from gpaw.test import equal
from gpaw.xc.rpa import RPACorrelation
from gpaw.xc.fxc import FXCCorrelation

pytestmark = pytest.mark.skipif(
    world.size != 1 and not compiled_with_sl(),
    reason='world.size != 1 and not compiled_with_sl()')


@pytest.mark.slow
def test_ralda_ralda_energy_Si(in_tmp_dir):
    a0 = 5.43
    cell = bulk('Si', 'fcc', a=a0).get_cell()
    Si = Atoms('Si2', cell=cell, pbc=True,
               scaled_positions=((0, 0, 0), (0.25, 0.25, 0.25)))

    kpts = monkhorst_pack((2, 2, 2))
    kpts += np.array([1 / 4, 1 / 4, 1 / 4])

    calc = GPAW(mode='pw',
                kpts=kpts,
                occupations=FermiDirac(0.001),
                communicator=serial_comm)
    Si.calc = calc
    Si.get_potential_energy()
    calc.diagonalize_full_hamiltonian(nbands=50)

    rpa = RPACorrelation(calc)
    E_rpa1 = rpa.calculate(ecut=[25, 50])

    fxc = FXCCorrelation(calc, xc='RPA', nlambda=16)
    E_rpa2 = fxc.calculate(ecut=[25, 50])

    fxc = FXCCorrelation(calc, xc='rALDA', unit_cells=[1, 1, 2])
    E_ralda = fxc.calculate(ecut=[25, 50])

    fxc = FXCCorrelation(calc, xc='rAPBE', unit_cells=[1, 1, 2])
    E_rapbe = fxc.calculate(ecut=[25, 50])

    fxc = FXCCorrelation(calc, xc='rALDA', av_scheme='wavevector')
    E_raldawave = fxc.calculate(ecut=[25, 50])

    fxc = FXCCorrelation(calc, xc='rAPBE', av_scheme='wavevector')
    E_rapbewave = fxc.calculate(ecut=[25, 50])

    fxc = FXCCorrelation(calc, xc='JGMs', av_scheme='wavevector',
                         Eg=3.1, nlambda=2)
    E_JGMs = fxc.calculate(ecut=[25, 50])

    fxc = FXCCorrelation(calc, xc='CP_dyn', av_scheme='wavevector',
                         nfrequencies=2, nlambda=2)
    E_CPdyn = fxc.calculate(ecut=[25, 50])

    equal(E_rpa1[-1], E_rpa2[-1], 0.01)
    equal(E_rpa2[-1], -12.6495, 0.002)
    equal(E_ralda[-1], -11.368, 0.002)
    equal(E_rapbe[-1], -11.149, 0.002)
    equal(E_raldawave[-1], -11.0910, 0.002)
    equal(E_rapbewave[-1], -10.831, 0.002)
    equal(E_JGMs[-1], -11.2561, 0.001)
    equal(E_CPdyn[-1], -7.8640, 0.001)



from ase.build import molecule
from ase.visualize import view
from gpaw import GPAW
from gpaw.wannier import calculate_overlaps

calc = GPAW(nbands=4)

atoms = molecule('H2O')
atoms.center(vacuum=3.)
atoms.calc = calc
atoms.get_potential_energy()

# Initialize the Wannier class
w = calculate_overlaps(calc, nwannier=4).localize_er()

view(w.centers_as_atoms())



from ase.build import fcc111
atoms = fcc111('Al', size=(1, 1, 2))
atoms.center(vacuum=4.0, axis=2)

from gpaw import GPAW
calc = GPAW(mode='pw',
            kpts=(4, 4, 1),
            symmetry='off',
            txt='al111.txt')
atoms.calc = calc
energy = atoms.get_potential_energy()
calc.write('al111.gpw', 'all')



# Creates: na2_casida_Ffe.png, na2_casida_Frho.png, na2_casida_Fphi.png
# -*- coding: utf-8 -*-
from gpaw.mpi import world

import numpy as np
import matplotlib.pyplot as plt

from gpaw.inducedfield.inducedfield_base import BaseInducedField
from gpaw.tddft.units import aufrequency_to_eV

assert world.size == 1, 'Script should be run in serial mode (one process).'


# Helper function
def do_plot(d_g, ng, box, atoms):
    # Take slice of data array
    d_yx = d_g[:, ng[1] // 2, :]
    y = np.linspace(0, box[0], ng[0])
    ylabel = u'x / Å'
    x = np.linspace(0, box[2], ng[2])
    xlabel = u'z / Å'

    # Plot
    plt.figure()
    ax = plt.subplot(1, 1, 1)
    X, Y = np.meshgrid(x, y)
    plt.contourf(X, Y, d_yx, 40)
    plt.colorbar()
    for atom in atoms:
        pos = atom.position
        plt.scatter(pos[2], pos[0], s=50, c='k', marker='o')
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.xlim([x[0], x[-1]])
    plt.ylim([y[0], y[-1]])
    ax.set_aspect('equal')


# Read InducedField object
ind = BaseInducedField('na2_casida_field.ind', readmode='all')

# Choose array
w = 1                                      # Frequency index
freq = ind.omega_w[w] * aufrequency_to_eV  # Frequency
box = np.diag(ind.atoms.get_cell())        # Calculation box
d_g = ind.Ffe_wg[w]                        # Data array
ng = d_g.shape                             # Size of grid
atoms = ind.atoms                          # Atoms

do_plot(d_g, ng, box, atoms)
plt.title(f'Field enhancement @ {freq:.2f} eV')
plt.savefig('na2_casida_Ffe.png', bbox_inches='tight')

# Imaginary part of density
d_g = ind.Frho_wg[w].imag * 1e-3  # Multiply by kick strength
ng = d_g.shape
do_plot(d_g, ng, box, atoms)
plt.title(f'Imaginary part of induced charge density @ {freq:.2f} eV')
plt.savefig('na2_casida_Frho.png', bbox_inches='tight')

# Imaginary part of potential
d_g = ind.Fphi_wg[w].imag * 1e-3  # Multiply by kick strength
ng = d_g.shape
do_plot(d_g, ng, box, atoms)
plt.title(f'Imaginary part of induced potential @ {freq:.2f} eV')
plt.savefig('na2_casida_Fphi.png', bbox_inches='tight')



from myqueue.task import task


def create_tasks():
    return [
        task('gs_MoS2.py@16:25m'),
        task('gs_WSe2.py@16:25m'),
        task('bb_MoS2.py@16:20h', deps='gs_MoS2.py'),
        task('bb_WSe2.py@16:20h', deps='gs_WSe2.py'),
        task('interpolate_bb.py', deps='bb_MoS2.py,bb_WSe2.py'),
        task('interlayer.py', deps='interpolate_bb.py')]



from myqueue.task import task


def create_tasks():
    tasks = [task('gaas.py+1@8:1h'),
             task('gaas.py+2@8:1h'),
             task('gaas.py+3@24:2h'),
             task('gaas.py+4@48:24h'),
             task('electrostatics.py@8:1:15m', deps=['gaas.py+1',
                                                     'gaas.py+2',
                                                     'gaas.py+3',
                                                     'gaas.py+4']),
             task('plot_energies.py', deps='electrostatics.py'),
             task('plot_potentials.py', deps='electrostatics.py'),
             task('BN.py+1@8:1h'),
             task('BN.py+2@8:1h'),
             task('BN.py+3@8:2h'),
             task('BN.py+4@24:2h'),
             task('BN.py+5@48:2h'),
             task('electrostatics_BN.py@1:15m', deps=['BN.py+1',
                                                      'BN.py+2',
                                                      'BN.py+3',
                                                      'BN.py+4',
                                                      'BN.py+5']),
             task('plot_energies_BN.py', deps='electrostatics_BN.py'),
             task('plot_potentials_BN.py', deps='electrostatics_BN.py'),
             task('plot_epsilon.py', deps='electrostatics_BN.py')]
    return tasks



# Creates: Ec_rpa.png
from myqueue.task import task


def create_tasks():
    return [task('si.groundstate.py'),
            task('si.range_rpa.py@8:30m', deps='si.groundstate.py'),
            task('si.compare.py', deps='si.range_rpa.py'),
            task('plot_ec.py', deps='si.range_rpa.py')]



import pytest
from gpaw.mpi import world
from gpaw.utilities import compiled_with_sl
import numpy as np
from ase.build import bulk
from gpaw import GPAW, FermiDirac, PW
from gpaw.response.g0w0 import G0W0
from gpaw.test import equal

pytestmark = pytest.mark.skipif(
    world.size != 1 and not compiled_with_sl(),
    reason='world.size != 1 and not compiled_with_sl()')


def test_response_gw0_hBN(in_tmp_dir):
    atoms = bulk('BN', 'zincblende', a=3.615)

    calc = GPAW(mode=PW(400),
                kpts={'size': (2, 2, 2), 'gamma': True},
                xc='LDA',
                eigensolver='rmm-diis',
                parallel={'domain': 1},
                occupations=FermiDirac(0.001))

    atoms.calc = calc
    atoms.get_potential_energy()

    calc.diagonalize_full_hamiltonian(scalapack=True)
    calc.write('BN_bulk_k2_ecut400_allbands.gpw', mode='all')

    gw = G0W0('BN_bulk_k2_ecut400_allbands.gpw',
              bands=(3, 5),
              nbands=9,
              nblocks=1,
              method='GW0',
              maxiter=5,
              ecut=40)

    result = gw.calculate()

    gaps = [3.28, 4.77, 4.97, 4.98, 4.98, 4.98]

    for i in range(result['iqp'].shape[0]):
        equal(np.min(result['iqp'][i, 0, :, 1]) -
              np.max(result['iqp'][i, 0, :, 0]), gaps[i], 0.03)



from ase import Atoms
from gpaw import GPAW

m = Atoms(symbols='C2',
          magmoms=[-0.5, 0.5],
          positions=[
              [0., 0., -0.62000006],
              [0., 0., 0.62000006]])
m.center(vacuum=4.0)

calc = GPAW(h=0.18,
            xc='PBE',
            basis='dzp',
            occupations={'name': 'fermi-dirac', 'width': 0.0},
            txt='C2_conv2.txt')

m.calc = calc
e2 = m.get_potential_energy()



import numpy as np

from ase.units import Bohr, Hartree

from gpaw import GPAW
from gpaw.external import ConstantElectricField
from gpaw.lcaotddft.hamiltonian import TimeDependentHamiltonian
from gpaw.lcaotddft.logger import TDDFTLogger
from gpaw.lcaotddft.propagators import create_propagator
from gpaw.tddft.units import attosec_to_autime


class LCAOTDDFT(GPAW):
    def __init__(self, filename=None, propagator=None, scale=None,
                 fxc=None, td_potential=None, **kwargs):
        self.time = 0.0
        self.niter = 0
        # TODO: deprecate kick keywords (and store them as td_potential)
        self.kick_strength = np.zeros(3)
        self.kick_ext = None
        self.tddft_initialized = False
        self.action = None
        tdh = TimeDependentHamiltonian(fxc=fxc, td_potential=td_potential,
                                       scale=scale)
        self.td_hamiltonian = tdh

        self.propagator = propagator
        if filename is None:
            kwargs['mode'] = kwargs.get('mode', 'lcao')
        self.default_parameters = GPAW.default_parameters.copy()
        self.default_parameters['symmetry'] = {'point_group': False}
        GPAW.__init__(self, filename, **kwargs)

        # Restarting from a file
        if filename is not None:
            # self.initialize()
            self.set_positions()

    def _write(self, writer, mode):
        GPAW._write(self, writer, mode)
        if self.tddft_initialized:
            w = writer.child('tddft')
            w.write(time=self.time,
                    niter=self.niter,
                    kick_strength=self.kick_strength,
                    propagator=self.propagator.todict())
            self.td_hamiltonian.write(w.child('td_hamiltonian'))

    def read(self, filename):
        reader = GPAW.read(self, filename)
        if 'tddft' in reader:
            r = reader.tddft
            self.time = r.time
            self.niter = r.niter
            self.kick_strength = r.kick_strength
            if self.propagator is None:
                self.propagator = r.propagator
            else:
                self.log('Note! Propagator changed!')
            self.td_hamiltonian.wfs = self.wfs
            self.td_hamiltonian.read(r.td_hamiltonian)

    def tddft_init(self):
        if self.tddft_initialized:
            return

        self.log('-----------------------------------')
        self.log('Initializing time-propagation TDDFT')
        self.log('-----------------------------------')
        self.log()

        assert self.wfs.dtype == complex

        self.timer.start('Initialize TDDFT')

        # Initialize Hamiltonian
        self.td_hamiltonian.initialize(self)

        # Initialize propagator
        self.propagator = create_propagator(self.propagator)
        self.propagator.initialize(self)

        self.log('Propagator:')
        self.log(self.propagator.get_description())
        self.log()

        # Add logger
        TDDFTLogger(self)

        # Call observers before propagation
        self.action = 'init'
        self.call_observers(self.niter)

        self.tddft_initialized = True
        self.timer.stop('Initialize TDDFT')

    def absorption_kick(self, kick_strength):
        self.tddft_init()

        self.timer.start('Kick')

        self.kick_strength = np.array(kick_strength, dtype=float)
        magnitude = np.sqrt(np.sum(self.kick_strength**2))
        direction = self.kick_strength / magnitude

        self.log('----  Applying absorption kick')
        self.log('----  Magnitude: %.8f Hartree/Bohr' % magnitude)
        self.log('----  Direction: %.4f %.4f %.4f' % tuple(direction))

        # Create hamiltonian object for absorption kick
        cef = ConstantElectricField(magnitude * Hartree / Bohr, direction)

        # Propagate kick
        self.propagator.kick(cef, self.time)

        # Call observers after kick
        self.action = 'kick'
        self.call_observers(self.niter)
        self.niter += 1
        self.timer.stop('Kick')

    def kick(self, ext):
        self.tddft_init()

        self.timer.start('Kick')

        self.log('----  Applying kick')
        self.log('----  %s' % ext)

        self.kick_ext = ext

        # Propagate kick
        self.propagator.kick(ext, self.time)

        # Call observers after kick
        self.action = 'kick'
        self.call_observers(self.niter)
        self.niter += 1
        self.timer.stop('Kick')

    def propagate(self, time_step=10, iterations=2000):
        self.tddft_init()

        time_step *= attosec_to_autime
        self.maxiter = self.niter + iterations

        self.log('----  About to do %d propagation steps' % iterations)

        self.timer.start('Propagate')
        while self.niter < self.maxiter:
            # Propagate one step
            self.time = self.propagator.propagate(self.time, time_step)

            # Call registered callback functions
            self.action = 'propagate'
            self.call_observers(self.niter)

            self.niter += 1
        self.timer.stop('Propagate')

    def replay(self, **kwargs):
        self.propagator = create_propagator(**kwargs)
        self.tddft_init()
        self.propagator.control_paw(self)



import numpy as np
from ase.build import molecule
from ase.parallel import parprint
from gpaw import GPAW
from gpaw.elf import ELF
from gpaw.test import equal
from gpaw.mpi import rank, world


def test_utilities_elf():
    atoms = molecule('CO')
    atoms.center(2.0)

    txt = '-'
    # txt = None

    calc = GPAW(h=0.24, txt=txt)
    atoms.calc = calc
    energy = atoms.get_potential_energy()

    elf = ELF(calc)
    elf.update()
    elf_G = elf.get_electronic_localization_function(gridrefinement=1)
    elf_g = elf.get_electronic_localization_function(gridrefinement=2)

    nt_G = calc.density.nt_sG[0]
    taut_G = elf.taut_sG[0]
    nt_grad2_G = elf.nt_grad2_sG[0]
    nt_grad2_g = elf.nt_grad2_sg[0]

    # integrate the CO bond
    if rank == 0:
        # bond area
        x0 = (atoms.positions[0][0] - 1.0) / atoms.get_cell()[0, 0]
        x1 = 1 - x0
        y0 = (atoms.positions[0][1] - 1.0) / atoms.get_cell()[1, 1]
        y1 = 1 - y0
        z0 = atoms.positions[1][2] / atoms.get_cell()[2, 2]
        z1 = atoms.positions[0][2] / atoms.get_cell()[2, 2]
        gd = calc.wfs.gd
        Gx0, Gx1 = int(gd.N_c[0] * x0), int(gd.N_c[0] * x1)
        Gy0, Gy1 = int(gd.N_c[1] * y0), int(gd.N_c[1] * y1)
        Gz0, Gz1 = int(gd.N_c[2] * z0), int(gd.N_c[2] * z1)
        finegd = calc.density.finegd
        gx0, gx1 = int(finegd.N_c[0] * x0), int(finegd.N_c[0] * x1)
        gy0, gy1 = int(finegd.N_c[1] * y0), int(finegd.N_c[1] * y1)
        gz0, gz1 = int(finegd.N_c[2] * z0), int(finegd.N_c[2] * z1)
        int1 = elf_G[Gx0:Gx1, Gy0:Gy1, Gz0:Gz1].sum() * gd.dv
        int2 = elf_g[gx0:gx1, gy0:gy1, gz0:gz1].sum() * finegd.dv
        parprint("Ints", int1, int2)
        parprint("Min, max G", np.min(elf_G), np.max(elf_G))
        parprint("Min, max g", np.min(elf_g), np.max(elf_g))
    #   The tested values (< r7887) do not seem to be correct
    #    equal(int1, 14.8078, 0.0001)
    #    equal(int2, 13.0331, 0.0001)

    # check spin-polarized version
    calc = GPAW(h=0.24,
                txt=txt,
                spinpol=True,
                parallel={'domain': world.size})
    atoms.calc = calc
    energy_spinpol = atoms.get_potential_energy()

    def check_diff(g1, g2, gd, txt):
        # print rank, txt, "shapes", g1.shape, g2.shape
        intd = gd.integrate(np.abs(g1 - g2))
        parprint(txt, 'integrated diff=', intd, end='')
        maxd = np.max(np.abs(g1 - g2))
        parprint('max diff=', maxd)
        equal(intd, 0, 1e-8)
        equal(maxd, 0, 1.e-9)

    nt_spinpol_G = calc.density.nt_sG.sum(axis=0)
    check_diff(nt_G, nt_spinpol_G, elf.finegd, 'nt_G')

    equal(energy, energy_spinpol, 0.0001)

    elf_spinpol = ELF(calc)
    elf_spinpol.update()
    elf_spinpol_G = elf_spinpol.get_electronic_localization_function(
        gridrefinement=1)
    elf_spinpol_g = elf_spinpol.get_electronic_localization_function(
        gridrefinement=2)
    taut_spinpol_G = elf_spinpol.taut_sG.sum(axis=0)
    check_diff(taut_G, taut_spinpol_G, elf.gd, 'taut_G')

    nt_grad2_spinpol_G = 2 * elf_spinpol.nt_grad2_sG.sum(axis=0)
    check_diff(nt_grad2_G, nt_grad2_spinpol_G, elf.gd, 'nt_grad2_G')

    nt_grad2_spinpol_g = 2 * elf_spinpol.nt_grad2_sg.sum(axis=0)
    check_diff(nt_grad2_g, nt_grad2_spinpol_g, elf.finegd, 'nt_grad2_g')

    check_diff(elf_G, elf_spinpol_G, elf.gd, 'elf_G')
    check_diff(elf_g, elf_spinpol_g, elf.finegd, 'elf_g')



# Creates: pdos.png
from gpaw import GPAW, restart
import matplotlib.pyplot as plt

# Density of States
plt.subplot(211)
slab, calc = restart('top.gpw')
e, dos = calc.get_dos(spin=0, npts=2001, width=0.2)
e_f = calc.get_fermi_level()
plt.plot(e - e_f, dos)
plt.axis([-15, 10, None, 4])
plt.ylabel('DOS')

molecule = range(len(slab))[-2:]

plt.subplot(212)
c_mol = GPAW('CO.gpw')
for n in range(2, 7):
    print('Band', n)
    # PDOS on the band n
    wf_k = [kpt.psit_nG[n] for kpt in c_mol.wfs.kpt_u]
    P_aui = [[kpt.P_ani[a][n] for kpt in c_mol.wfs.kpt_u]
             for a in range(len(molecule))]
    e, dos = calc.get_all_electron_ldos(mol=molecule, spin=0, npts=2001,
                                        width=0.2, wf_k=wf_k, P_aui=P_aui)
    plt.plot(e - e_f, dos, label='Band: ' + str(n))
plt.legend()
plt.axis([-15, 10, None, None])
plt.xlabel('Energy [eV]')
plt.ylabel('All-Electron PDOS')
plt.savefig('pdos.png')
plt.show()



from gpaw import GPAW
from gpaw.lrtddft import LrTDDFT

c = GPAW('Be_gs_8bands.gpw')

istart = 0  # band index of the first occ. band to consider
jend = 8  # band index of the last unocc. band to consider
lr = LrTDDFT(c, xc='LDA', restrict={'istart': istart, 'jend': jend},
             nspins=2)  # force the calculation of triplet excitations also
lr.write('lr.dat.gz')



# Creates: energies.png
import numpy as np
import matplotlib.pyplot as plt

data = np.load('formation_energies.npz')

repeat = data['repeats']
uncorrected = data['uncorrected']
inv_repeat = 1 / repeat[1:]
line = np.polyfit(inv_repeat, uncorrected[1:].real, deg=1)
f = np.poly1d(line)
points = np.linspace(0, 1, 100)
corrected = data['corrected']
plt.plot(1 / repeat, uncorrected.real, 'o', label='No corrections')
plt.plot(1 / repeat, corrected.real, 'p', label='FNV corrected')
plt.plot(points, f(points), "--", color='C0')
plt.axhline(line[1], linestyle="dashed", color='C1')

plt.xlabel('Supercell size', fontsize=18)
plt.ylabel('Energy difference (eV)', fontsize=18)
plt.xlim(left=0.06)
plt.xticks(1 / repeat, [str(x) for x in repeat])
plt.legend(loc='lower left')
plt.savefig('energies.png', bbox_inches='tight', dpi=300)



import numpy as np

from ase import Atoms
from gpaw import GPAW, FermiDirac, Davidson
from gpaw.test import equal


def test_fermilevel(in_tmp_dir):
    calc = GPAW(nbands=1,
                eigensolver=Davidson(6),
                occupations=FermiDirac(0.0))
    atoms = Atoms('He', pbc=True, calculator=calc)
    atoms.center(vacuum=3)

    _ = atoms.get_potential_energy()
    assert np.isinf(calc.get_fermi_level())
    calc.set(nbands=3, convergence={'bands': 2})
    atoms.get_potential_energy()
    homo, lumo = calc.get_homo_lumo()
    equal(homo, -15.4473, 0.01)
    equal(lumo, -0.2566, 0.01)
    calc.write('test')
    assert np.all(GPAW('test', txt=None).get_homo_lumo() == (homo, lumo))
    ef = calc.get_fermi_level()
    equal(ef, -7.85196, 0.01)

    calc.set(occupations=FermiDirac(0.1))
    _ = atoms.get_potential_energy()
    ef = calc.get_fermi_level()
    equal(ef, -7.85196, 0.01)
    calc.write('test')
    equal(GPAW('test', txt=None).get_fermi_level(), ef, 1e-8)



scalapack = False

extra_compile_args = ['-O3', '-std=c99', '-fpic']

compiler = 'gcc'
mpicompiler = '/home/firegam/CAMd/openmpi-1.4.3-1/bin/mpicc'
mpilinker = mpicompiler

libraries = ['mkl_lapack', 'mkl_core', 'mkl_sequential', 'mkl_gf_lp64', 'iomp5']

mkl_lib_path = '/opt/intel/Compiler/11.1/072/mkl/lib/em64t/'
ompi_lib_path = '/home/firegam/CAMd/openmpi-1.4.3-1/lib'

library_dirs = [mkl_lib_path, ompi_lib_path]

extra_link_args =['-Wl,-rpath='+mkl_lib_path+',-rpath='+ompi_lib_path]

define_macros += [('GPAW_NO_UNDERSCORE_CBLACS', '1')]
define_macros += [('GPAW_NO_UNDERSCORE_CSCALAPACK', '1')]



def create_tasks():
    from myqueue.task import task
    return [
        task('HAl100.py'),
        task('stm.agts.py', deps='HAl100.py')]


if __name__ == '__main__':
    import sys
    sys.argv = ['', 'HAl100.gpw']
    exec(open('stm.py').read())



import numpy as np

pos = np.asarray([1.450 * a for a in range(6)])
n = len(pos)
d2_ii = np.zeros((n, n))
for i1, i2 in np.ndindex(n, n):
    d2_ii[i1, i2] = ((pos[i1] - pos[i2])**2).sum()

d_ii = np.sqrt(d2_ii)
U = 11.26
t_m = -2.4
t_lm = -1.0
# Two particle interaction
V = U / np.sqrt(1.0 + 0.6117 * d2_ii)
V[0, -1] = V[-1, 0] = 0.0
V[0, 0] = V[-1, -1] = 0.0
# One particle part
mask_c = d_ii < 1.5
mask_c *= d_ii > 0.0
mask_c = mask_c.astype(int)
nbf = len(d_ii)
h = t_m * np.ones((nbf, nbf)) * mask_c
h[0, 1] = h[1, 0] = t_lm
h[-2, -1] = h[-1, -2] = t_lm

# H_lead
h1 = np.zeros((2, 2), complex)
t_l = -20.0
h1[0, 1] = t_l
h1[1, 0] = t_l
nbf = len(h)

# H_scat
H = np.zeros((nbf + 2, nbf + 2), complex)
H[0, 1] = H[1, 0] = t_l
H[-2, -1] = H[-1, -2] = t_l
H[1:-1, 1:-1] = h

# Hartree potential of the ions (Z=1)
ion_shift = np.zeros((n, n))
for i in range(n):
    ion_shift[i, i] += -0.5 * V[i, i]
    for k in range(n):
        if k != i:
            ion_shift[i, i] += -1.0 * V[i, k]

H = H.astype(complex)
V = V.astype(complex)
ion_shift = ion_shift.astype(complex)

if __name__ == '__main__':
    print(np.around(H.real, 2))
    print(np.around(V.real, 2))
    print(np.around(ion_shift.diagonal().real, 2))



from pathlib import Path
from gpaw.mpi import world
from gpaw.response.df import DielectricFunction
from gpaw.response.qeh import BuildingBlock

df = DielectricFunction(calc='MoS2_gs_fulldiag.gpw',
                        eta=0.001,
                        domega0=0.05,
                        omega2=10.0,
                        nblocks=8,
                        ecut=150,
                        truncation='2D')

buildingblock = BuildingBlock('MoS2', df, qmax=3.0)

buildingblock.calculate_building_block()

if world.rank == 0:
    Path('MoS2_gs_fulldiag.gpw').unlink()



import ase.db
from ase.build import bulk
import numpy as np
from gpaw.hybrids.energy import non_self_consistent_energy as nsc_energy
from gpaw import GPAW, PW

a0 = 5.43

con = ase.db.connect('si.db')

for k in range(2, 9):
    for a in np.linspace(a0 - 0.04, a0 + 0.04, 5):
        id = con.reserve(a=a, k=k)
        if id is None:
            continue
        si = bulk('Si', 'diamond', a)
        si.calc = GPAW(kpts=(k, k, k),
                       mode=PW(400),
                       xc='PBE',
                       eigensolver='rmm-diis',
                       txt=None)
        si.get_potential_energy()
        name = f'si-{a:.2f}-{k}'
        si.calc.write(name + '.gpw', mode='all')
        epbe0 = nsc_energy(name + '.gpw', 'PBE0').sum()

        con.write(si, a=a, k=k, epbe0=epbe0)
        del con[id]



from ase.build import molecule
from ase.visualize import view
from gpaw import GPAW
from gpaw.wannier import calculate_overlaps

calc = GPAW(nbands=5)
atoms = molecule('CO')
atoms.center(vacuum=3.)
atoms.calc = calc
atoms.get_potential_energy()

# Initialize the Wannier class
w = calculate_overlaps(calc, nwannier=5).localize_er()
view(w.centers_as_atoms())



# Creates: C_GW.png
import pickle
import numpy as np
from ase.parallel import paropen
import matplotlib.pyplot as plt

plt.figure(figsize=(6.5, 4.5))

ecuts = np.array([100, 200, 300, 400])
color = ['ro-', 'bo-', 'go-', 'ko-', 'co-', 'mo-', 'yo-']
direct_gap = np.zeros(4)

for j, k in enumerate([6, 8, 10, 12]):
    for i, ecut in enumerate([100, 200, 300, 400]):
        fil = pickle.load(
            paropen(f'C-g0w0_k{k}_ecut{ecut}_results.pckl', 'rb'))
        direct_gap[i] = fil['qp'][0, 0, 1] - fil['qp'][0, 0, 0]
    plt.plot(ecuts, direct_gap, color[j],
             label=f'({k}x{k}x{k}) k-points')

plt.xlabel('Cutoff energy (eV)')
plt.ylabel('Direct band gap (eV)')
plt.title('non-selfconsistent G0W0@LDA')
plt.legend(loc='lower right')
plt.savefig('C_GW.png')
plt.show()



from gpaw.tddft import TDDFT
from gpaw.inducedfield.inducedfield_tddft import TDDFTInducedField

# Load TDDFT object
td_calc = TDDFT('na2_td.gpw')

# Load and attach InducedField object
ind = TDDFTInducedField(filename='na2_td.ind',
                        paw=td_calc,
                        restart_file='na2_td.ind')

# Continue propagation as usual
time_step = 20.0
iterations = 250
td_calc.propagate(time_step, iterations, 'na2_td_dm.dat', 'na2_td.gpw')

# Save TDDFT and InducedField objects
td_calc.write('na2_td.gpw', mode='all')
ind.write('na2_td.ind')



"""GLLBSC band gap tutorial

Calculate the band gap of Si using GLLBSC.
"""
# P1
from ase.build import bulk
from gpaw import GPAW, PW, FermiDirac

# Ground state calculation
atoms = bulk('Si', 'diamond', 5.431)
calc = GPAW(mode=PW(200),
            xc='GLLBSC',
            kpts=(7, 7, 7),  # Choose and converge carefully!
            occupations=FermiDirac(0.01),
            txt='gs.out')
atoms.calc = calc
atoms.get_potential_energy()
# P2
# Calculate the discontinuity potential and the discontinuity
homo, lumo = calc.get_homo_lumo()
response = calc.hamiltonian.xc.response
dxc_pot = response.calculate_discontinuity_potential(homo, lumo)
KS_gap, dxc = response.calculate_discontinuity(dxc_pot)

# Fundamental band gap = Kohn-Sham band gap + derivative discontinuity
QP_gap = KS_gap + dxc

print(f'Kohn-Sham band gap:         {KS_gap:.2f} eV')
print(f'Discontinuity from GLLB-sc: {dxc:.2f} eV')
print(f'Fundamental band gap:       {QP_gap:.2f} eV')



from ase.build import molecule
from ase.optimize import BFGS
from gpaw import GPAW, Davidson
from gpaw.mixer import MixerDif

for name in ['H2', 'N2', 'O2', 'NO']:
    mol = molecule(name)
    mol.center(vacuum=5.0)
    calc = GPAW(xc='PBE',
                h=0.2,
                eigensolver=Davidson(3),
                txt=name + '.txt',
                convergence={'eigenstates': 1e-10})
    if name == 'NO':
        mol.translate((0, 0.1, 0))
        calc.set(mixer=MixerDif(0.05, 5))
    mol.calc = calc

    opt = BFGS(mol, logfile=name + '.log', trajectory=name + '.traj')
    opt.run(fmax=0.05)
    calc.write(name + '.gpw')



"""Calculate dipole matrix elements."""

from math import pi
from typing import List, Iterable

from ase.units import Bohr
import numpy as np

from gpaw import GPAW
from gpaw.grid_descriptor import GridDescriptor
from gpaw.setup import Setup
from gpaw.mpi import serial_comm
from gpaw.hints import Array2D, Array3D, Array4D


def dipole_matrix_elements(gd: GridDescriptor,
                           psit_nR: List[Array3D],
                           P_nI: Array2D,
                           position_av: Array2D,
                           setups: List[Setup],
                           center: Iterable[float]) -> Array3D:
    """Calculate dipole matrix-elements.

    gd:
        Grid-descriptor.
    psit_nG:
        Wave functions in atomic units.
    P_nI:
        PAW projections.
    setups:
        PAW setups.

    Returns matrix elements in atomic units.
    """
    assert gd.comm.size == 1
    dipole_nnv = np.empty((len(psit_nR), len(psit_nR), 3))

    for na, psita_R in enumerate(psit_nR):
        for nb, psitb_R in enumerate(psit_nR[:na + 1]):
            d_v = -gd.dipole_moment(psita_R * psitb_R, center)
            dipole_nnv[na, nb] = d_v
            dipole_nnv[nb, na] = d_v

    scenter_c = np.linalg.solve(gd.cell_cv.T, center)
    spos_ac = np.linalg.solve(gd.cell_cv.T, position_av.T).T % 1.0
    spos_ac -= scenter_c - 0.5
    spos_ac %= 1.0
    spos_ac += scenter_c - 0.5
    position_av = spos_ac.dot(gd.cell_cv)

    R_aiiv = []
    for setup, position_v in zip(setups, position_av):
        Delta_iiL = setup.Delta_iiL
        R_iiv = Delta_iiL[:, :, [3, 1, 2]] * (4 * pi / 3)**0.5
        R_iiv += position_v * setup.Delta_iiL[:, :, :1] * (4 * pi)**0.5
        R_aiiv.append(R_iiv)

    I1 = 0
    for R_iiv in R_aiiv:
        I2 = I1 + len(R_iiv)
        P_ni = P_nI[:, I1:I2]
        dipole_nnv += np.einsum('mi, ijv, nj -> mnv', P_ni, R_iiv, P_ni)
        I1 = I2

    return dipole_nnv


def dipole_matrix_elements_from_calc(calc: GPAW,
                                     n1: int,
                                     n2: int,
                                     center: Iterable[float] = None
                                     ) -> List[Array4D]:
    """Calculate dipole matrix-elements (units: Å)."""
    wfs = calc.wfs
    assert wfs.kd.gamma

    gd = wfs.gd.new_descriptor(comm=serial_comm)
    position_av = calc.atoms.positions / Bohr

    if center is None:
        center_v = gd.cell_cv.sum(axis=0) / 2
    else:
        center_v = np.asarray(center) / Bohr

    d_snnv = []
    for spin in range(calc.get_number_of_spins()):
        psit_nR = [calc.get_pseudo_wave_function(band=n, spin=spin,
                                                 pad=False) * Bohr**1.5
                   for n in range(n1, n2)]
        projections = wfs.kpt_qs[0][spin].projections.view(n1, n2)
        P_nI = projections.collect()

        if wfs.world.rank == 0:
            d_nnv = dipole_matrix_elements(gd,
                                           psit_nR,
                                           P_nI,
                                           position_av,
                                           wfs.setups,
                                           center_v) * Bohr
            d_snnv.append(d_nnv)
    return d_snnv


def main(argv: List[str] = None) -> None:
    import argparse

    parser = argparse.ArgumentParser(
        prog='python3 -m gpaw.utilities.dipole',
        description='Calculate dipole matrix elements.  Units: Å.')

    add = parser.add_argument

    add('file', metavar='input-file',
        help='GPW-file with wave functions.')
    add('-n', '--band-range', nargs=2, type=int, default=[0, 0],
        metavar=('n1', 'n2'), help='Include bands: n1 <= n < n2.')
    add('-c', '--center', metavar='x,y,z',
        help='Center of charge distribution.  Default is middle of unit '
        'cell.')

    if hasattr(parser, 'parse_intermixed_args'):
        args = parser.parse_intermixed_args(argv)
    else:
        args = parser.parse_args(argv)

    calc = GPAW(args.file)

    n1, n2 = args.band_range
    nbands = calc.get_number_of_bands()
    n2 = n2 or n2 + nbands

    if args.center:
        center = [float(x) for x in args.center.split(',')]
    else:
        center = calc.atoms.cell.sum(axis=0) / 2  # center of cell

    d_snnv = dipole_matrix_elements_from_calc(calc, n1, n2, center)

    if calc.wfs.world.rank > 0:
        return

    print('Number of bands:', nbands)
    print('Number of valence electrons:', calc.get_number_of_electrons())
    print()

    for spin, d_nnv in enumerate(d_snnv):
        print(f'Spin={spin}')

        for direction, d_nn in zip('xyz', d_nnv.T):
            print(f' <{direction}>',
                  ''.join(f'{n:8}' for n in range(n1, n2)))
            for n in range(n1, n2):
                print(f'{n:4}', ''.join(f'{d:8.3f}' for d in d_nn[n - n1]))


if __name__ == '__main__':
    main()



from ase.build import bulk
from gpaw import GPAW, PW

a = 5.421
si = bulk('Si', 'fcc', a=a)
# or equivalently:
# b = a / 2
# from ase import Atoms
# si = Atoms('Si2', cell=[[0, b, b], [b, 0, b], [b, b, 0]], pbc=True,
#           scaled_positions=[[0, 0, 0], [0.25, 0.25, 0.25]])

for x in [100, 200, 300, 400, 500, 600, 700, 800]:
    # for x in [0.24, 0.22, 0.20, 0.18, 0.16, 0.14, 0.12, 0.1]:
    calc = GPAW(mode=PW(x),
                # h=x,
                xc='PBE',
                kpts=(4, 4, 4),
                txt=f'convergence_{x}.txt')

    si.calc = calc

    print(x, si.get_potential_energy())



# Creates: si-a.png
import matplotlib.pyplot as plt
import ase.db
from ase.eos import EquationOfState


def lattice_constant(volumes, energies):
    eos = EquationOfState(volumes, energies)
    v, e, B = eos.fit()
    a = (v * 4)**(1 / 3)
    return a
    
    
con = ase.db.connect('si.db')
results = []
K = list(range(2, 9))
A = []
A0 = []
for k in K:
    rows = list(con.select(k=k))
    V = [row.volume for row in rows]
    E = [row.energy for row in rows]
    E0 = [row.epbe0 for row in rows]
    A.append(lattice_constant(V, E))
    A0.append(lattice_constant(V, E0))
    
print(K, A, A0)

plt.plot(K, A, label='PBE')
plt.plot(K, A0, label='PBE0')
plt.xlabel('number of k-points')
plt.ylabel('lattice constant [Ang]')
plt.savefig('si-a.png')



# creates: pt_h2.png

from ase import Atoms
from ase.io import write

a = 2.41  # Pt binding length
b = 0.90  # H2 binding length
c = 1.70  # Pt-H binding length
L = 7.00  # width of unit cell

# Setup the Atoms for the scattering region.
atoms = Atoms('Pt5H2Pt5',
              cell=[3 * a + b + 2 * c, L, L],
              pbc=True)

atoms.positions[:5, 0] = [(i - 2.5) * a for i in range(5)]
atoms.positions[-5:, 0] = [(i - 2.5) * a + b + 2 * c for i in range(4, 9)]
atoms.positions[5:7, 0] = [1.5 * a + c, 1.5 * a + c + b]
atoms.positions[:, 1:] = L / 2.

write('pt_h2.pov',
      atoms,
      show_unit_cell=2,
      povray_settings={'transparent': False}).render()



from gpaw.lcaotddft import LCAOTDDFT
from gpaw.lcaotddft.dipolemomentwriter import DipoleMomentWriter
from gpaw import setup_paths

# Insert the path to the created basis set
setup_paths.insert(0, '.')

# Parallelzation settings
parallel = {'sl_auto': True, 'domain': 2, 'augment_grids': True}

# Time propagation
td_calc = LCAOTDDFT('gs.gpw', parallel=parallel, txt='td.out')
DipoleMomentWriter(td_calc, 'dm.dat')
td_calc.absorption_kick([1e-5, 0.0, 0.0])
td_calc.propagate(10, 3000)



import numpy as np
from scipy.special import erfcx

from ase.units import Ha


K_G = 8 * np.sqrt(2) / (3 * np.pi**2)  # 0.382106112167171


class Coefficients:
    r"""Coefficient calculator for GLLB functionals.

    This class implements the calculation of sqrt(E) coefficients as given by
    Eq. (16) of https://doi.org/10.1103/PhysRevB.82.115106.

    Parameters:

    eps (in eV):
        This parameter cuts sqrt(E) to zero for E < eps.
        The cut is supposed to help convergence with degenerate systems.
        This parameter should be small.
    width (in eV):
        If this parameter is set, then a smoothed variant of the sqrt(E)
        expression is used. This parameter sets the energy width of
        the smoothing.
        The eps parameter is ignored when the width parameter is set.
    metallic:
        If this parameter is set, then Fermi level is used as the reference
        energy in the sqrt(E) expression instead of the HOMO energy.
        This is necessary to get sensible results in metallic systems.
    """
    def __init__(self,
                 eps: float = 0.05,
                 width: float = None,
                 metallic: bool = False):
        self.eps = eps / Ha
        self.metallic = metallic
        if width is not None:
            width = width / Ha
            self.eps = None  # Make sure that eps is not used with width
        self.width = width

    def initialize(self, wfs):
        self.wfs = wfs

    def initialize_1d(self, ae):
        self.ae = ae

    def get_description(self):
        desc = []
        if self.eps is not None:
            desc += ['eps={:.4f} eV'.format(self.eps * Ha)]
        if self.metallic:
            desc += ['metallic']
        if self.width is not None:
            desc += ['width={:.4f} eV'.format(self.width * Ha)]
        return ', '.join(desc)

    def f(self, energy_n):
        """Calculate the sqrt(E)-like coefficient.

        See the class description for details.
        """
        w_n = np.zeros_like(energy_n, dtype=float)
        if self.width is None:
            flt_n = energy_n > self.eps
            w_n[flt_n] = np.sqrt(energy_n[flt_n])
        else:
            prefactor = 0.5 * np.sqrt(np.pi * self.width)
            rel_energy_n = energy_n / self.width
            # Evaluate positive energies
            flt_n = energy_n > 0.0
            w_n[flt_n] = (np.sqrt(energy_n[flt_n])
                          + prefactor * erfcx(np.sqrt(rel_energy_n[flt_n])))
            # Evaluate negative energies
            flt_n = np.logical_not(flt_n)
            w_n[flt_n] = prefactor * np.exp(rel_energy_n[flt_n])
        return w_n

    def get_reference_homo_1d(self):
        e_j = np.asarray(self.ae.e_j)
        f_j = np.asarray(self.ae.f_j)
        homo = np.max(e_j[f_j > 1e-3])
        return homo

    def get_reference_lumo_1d(self):
        e_j = np.asarray(self.ae.e_j)
        f_j = np.asarray(self.ae.f_j)
        lumo = np.min(e_j[f_j < 1e-3])
        return lumo

    def get_coefficients_1d(self, smooth=False):
        homo = self.get_reference_homo_1d()
        if smooth:
            w_ln = []
            for e_n, f_n in zip(self.ae.e_ln, self.ae.f_ln):
                e_n = np.asarray(e_n)
                f_n = np.asarray(f_n)
                w_n = f_n * K_G * self.f(homo - e_n)
                w_ln.append(w_n)
            return w_ln
        else:
            e_j = np.asarray(self.ae.e_j)
            f_j = np.asarray(self.ae.f_j)
            return f_j * K_G * self.f(homo - e_j)

    def get_coefficients_1d_for_lumo_perturbation(self, smooth=False):
        homo = self.get_reference_homo_1d()
        lumo = self.get_reference_lumo_1d()
        e_j = np.asarray(self.ae.e_j)
        f_j = np.asarray(self.ae.f_j)
        return f_j * K_G * (self.f(lumo - e_j) - self.f(homo - e_j))

    def get_reference_energies(self, homolumo=None):
        eref_s = []
        eref_lumo_s = []
        if self.metallic:
            # Use Fermi level as reference levels
            assert homolumo is None
            fermilevel = self.wfs.fermi_level
            assert isinstance(fermilevel, float), \
                'GLLBSCM supports only a single Fermi level'
            for s in range(self.wfs.nspins):
                eref_s.append(fermilevel)
                eref_lumo_s.append(fermilevel)
        elif homolumo is None:
            # Find homo and lumo levels for each spin
            for s in range(self.wfs.nspins):
                homo, lumo = self.wfs.get_homo_lumo(s)
                # Check that homo and lumo are reasonable
                if homo > lumo:
                    m = ("GLLBSC error: HOMO is higher than LUMO. "
                         "Are you using `xc='GLLBSC'` for a metallic system? "
                         "If yes, try using `xc='GLLBSCM'` instead.")
                    raise RuntimeError(m)

                eref_s.append(homo)
                eref_lumo_s.append(lumo)
        else:
            eref_s, eref_lumo_s = homolumo
            if not isinstance(eref_s, (list, tuple, np.ndarray)):
                eref_s = [eref_s]
                eref_lumo_s = [eref_lumo_s]
        return eref_s, eref_lumo_s

    def get_coefficients(self, kpt_u, homolumo=None):
        eref_s, eref_lumo_s = self.get_reference_energies(homolumo=homolumo)
        w_kn = []
        for kpt in kpt_u:
            w_n = self.f(eref_s[kpt.s] - kpt.eps_n)
            w_n *= kpt.f_n * K_G
            w_kn.append(w_n)
        return w_kn

    def get_coefficients_for_lumo_perturbation(self, kpt_u, homolumo=None):
        eref_s, eref_lumo_s = self.get_reference_energies(homolumo=homolumo)
        w_kn = []
        for kpt in kpt_u:
            w_n = (self.f(eref_lumo_s[kpt.s] - kpt.eps_n)
                   - self.f(eref_s[kpt.s] - kpt.eps_n))
            w_n *= kpt.f_n * K_G
            w_kn.append(w_n)
        return w_kn



# Creates: geom.png
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap
from ase.units import Bohr
from ase.visualize.plot import plot_atoms

from gpaw.tddft import TDDFT


# Initialize TDDFT and QSFDTD
td_calc = TDDFT('gs.gpw')


def generate_xygrid(d, g, box):

    vslice = 2  # yx

    # Determine the array lengths in each dimension
    ng = d.shape

    X = None
    Y = None
    U = None
    V = None

    # Slice data
    d_slice = np.rollaxis(d, vslice)[g[vslice], :, :]
    d_proj = np.zeros(d_slice.shape)
    for ind, val in np.ndenumerate(d_slice):
        d_proj[ind] = np.where(
            np.append(
                np.rollaxis(d, vslice)[:, ind[0], ind[1]], 1.0) != 0)[0][0]

    # Grids
    x = np.linspace(0, box[1], ng[1])
    y = np.linspace(0, box[0], ng[0])

    # Meshgrid and corresponding data
    X, Y = np.meshgrid(x, y)
    U = np.real(d_slice[1])  # y
    V = np.real(d_slice[0])  # x

    # Spacing
    dx = x[1] - x[0]
    dy = y[1] - y[0]
    return d_slice, d_proj, (x, y, dx, dy), (X, Y, U, V)


poisson_solver = td_calc.hamiltonian.poisson
atoms = td_calc.atoms

box = np.diagonal(poisson_solver.cl.gd.cell_cv) * Bohr  # in Ang

# create figure
plt.figure(1, figsize=(4, 4))
plt.rcParams['font.size'] = 14

# prepare data
plotData = poisson_solver.classical_material.beta[0]
ng = plotData.shape

axis = 2
ax = plt.subplot(1, 1, 1)
g = [None, None, ng[2] // 2]

dmy1, d_proj, (x, y, dx, dy), dmy2 = generate_xygrid(plotData, g, box)

# choose the colourmap for the polarizable material here
plt.imshow(d_proj, interpolation='bicubic', origin='lower',
           cmap=ListedColormap(["goldenrod", "white"]),
           extent=[x[0] - dx / 2, x[-1] + dx / 2,
                   y[0] - dy / 2, y[-1] + dy / 2])

# Plot atoms
# switch x and y orientation for yx plot
pos = atoms.get_positions()
pos[:, [0, 1]] = pos[:, [1, 0]]
atoms.set_positions(pos)

cell = atoms.get_cell()
cell[:, [0, 1]] = cell[:, [1, 0]]
atoms.set_cell(cell)

# ASE plot atoms function
i, j = 1, 0

offset = np.array(
    [poisson_solver.qm.corner1[i],
     poisson_solver.qm.corner1[j]]) * 2 * Bohr

bbox = np.array(
    [poisson_solver.qm.corner1[i],
     poisson_solver.qm.corner1[j],
     poisson_solver.qm.corner2[i],
     poisson_solver.qm.corner2[j]]) * Bohr

plot_atoms(atoms, ax=None, show_unit_cell=2, offset=offset, bbox=bbox)

ax.autoscale()

# Classical grid
dmy1, dmy_proj, (x, y, dx, dy), dmy3 = generate_xygrid(plotData, g, box)
xx, yy = np.meshgrid(x, y)
plt.scatter(xx,
            yy,
            s=0.75, c='k', marker='o')

# Quantum grid
dmy1, dmy_proj, (x, y, dx, dy), dmy3 = generate_xygrid(
    plotData, g, box=np.diagonal(poisson_solver.qm.gd.cell_cv) * Bohr)
xx, yy = np.meshgrid(x, y)
plt.scatter(poisson_solver.qm.corner1[i] * Bohr + xx,
            poisson_solver.qm.corner1[j] * Bohr + yy,
            s=0.25, c='k', marker='o')

# Labels
plt.xlabel('y [Ang]')
plt.ylabel('x [Ang]')

# Plot
plt.tight_layout()
plt.savefig('geom.png')



from typing import Tuple, Dict, Sequence, List, Union
from pathlib import Path

import numpy as np
from ase import Atoms

from gpaw import GPAW
from gpaw.projections import Projections
from gpaw.utilities.partition import AtomPartition
from gpaw.setup import Setup
from gpaw.kpt_descriptor import KPointDescriptor
from gpaw.utilities.ibz2bz import construct_symmetry_operators
from .functions import WannierFunctions
from gpaw.hints import Array2D, Array3D, Array4D


class WannierOverlaps:
    def __init__(self,
                 atoms: Atoms,
                 nwannier: int,
                 monkhorst_pack_size: Sequence[int],
                 kpoints: Array2D,
                 fermi_level: float,
                 directions: Dict[Tuple[int, ...], int],
                 overlaps: Array4D,
                 projections: Array3D = None,
                 proj_indices_a: List[List[int]] = None):

        self.atoms = atoms
        self.nwannier = nwannier
        self.monkhorst_pack_size = tuple(monkhorst_pack_size)
        self.kpoints = kpoints
        self.fermi_level = fermi_level
        self.directions = directions

        self.nkpts, ndirs, self.nbands, nbands = overlaps.shape
        assert nbands == self.nbands
        assert self.nkpts == np.prod(monkhorst_pack_size)
        assert ndirs == len(directions)

        self._overlaps = overlaps
        self.projections = projections
        self.proj_indices_a = proj_indices_a

    def overlap(self,
                bz_index: int,
                direction: Tuple[int, ...]) -> Array2D:
        dindex = self.directions.get(direction)
        if dindex is not None:
            return self._overlaps[bz_index, dindex]

        size = self.monkhorst_pack_size
        i_c = np.unravel_index(bz_index, size)
        i2_c = np.array(i_c) + direction
        bz_index2 = np.ravel_multi_index(i2_c, size, 'wrap')
        direction2 = tuple([-d for d in direction])
        dindex2 = self.directions[direction2]
        return self._overlaps[bz_index2, dindex2].T.conj()

    def localize_er(self,
                    maxiter: int = 100,
                    tolerance: float = 1e-5,
                    verbose: bool = not False) -> WannierFunctions:
        from .edmiston_ruedenberg import localize
        return localize(self, maxiter, tolerance, verbose)

    def localize_w90(self,
                     prefix: str = 'wannier',
                     folder: Union[Path, str] = 'W90',
                     nwannier: int = None,
                     **kwargs) -> WannierFunctions:
        from .w90 import Wannier90
        w90 = Wannier90(prefix, folder)
        w90.write_input_files(overlaps=self, **kwargs)
        w90.run_wannier90()
        return w90.read_result()


def dict_to_proj_indices(dct: Dict[Union[int, str], str],
                         setups: List[Setup]) -> List[List[int]]:
    """Convert dict to lists of projector function indices.

    >>> from gpaw.setup import create_setup
    >>> setup = create_setup('Si')  # 3s, 3p, *s, *p, *d
    >>> setup.n_j
    [3, 3, -1, -1, -1]
    >>> setup.l_j
    [0, 1, 0, 1, 2]
    >>> dict_to_proj_indices({'Si': 'sp', 1: 's'}, [setup, setup])
    [[0, 1, 2, 3], [0]]
    """
    indices_a = []
    for a, setup in enumerate(setups):
        ll = dct.get(a, dct.get(setup.symbol, ''))
        indices = []
        i = 0
        for n, l in zip(setup.n_j, setup.l_j):
            if n > 0 and 'spdf'[l] in ll:
                indices += list(range(i, i + 2 * l + 1))
            i += 2 * l + 1
        indices_a.append(indices)
    return indices_a


def calculate_overlaps(calc: GPAW,
                       nwannier: int,
                       projections: Dict[Union[int, str], str] = None,
                       n1: int = 0,
                       n2: int = 0,
                       spinors: bool = False,
                       spin: int = 0) -> WannierOverlaps:
    """Create WannierOverlaps object from DFT calculation.
    """
    assert not spinors

    if n2 <= 0:
        n2 += calc.get_number_of_bands()

    bzwfs = BZRealSpaceWaveFunctions.from_calculation(calc, n1, n2, spin)

    proj_indices_a = dict_to_proj_indices(projections or {},
                                          calc.setups)

    offsets = [0]
    for indices in proj_indices_a:
        offsets.append(offsets[-1] + len(indices))
    nproj = offsets.pop()

    if projections is not None:
        assert nproj == nwannier

    kd = bzwfs.kd
    gd = bzwfs.gd
    size = kd.N_c

    icell = calc.atoms.cell.reciprocal()
    directions = {direction: i
                  for i, direction
                  in enumerate(find_directions(icell, size))}
    Z_kdnn = np.empty((kd.nbzkpts, len(directions), n2 - n1, n2 - n1), complex)

    spos_ac = calc.spos_ac
    setups = calc.wfs.setups

    proj_kmn = np.zeros((kd.nbzkpts, nproj, n2 - n1), complex)

    for bz_index1 in range(kd.nbzkpts):
        wf1 = bzwfs[bz_index1]
        i1_c = np.unravel_index(bz_index1, size)
        for direction, d in directions.items():
            i2_c = np.array(i1_c) + direction
            bz_index2 = np.ravel_multi_index(i2_c, size, 'wrap')
            wf2 = bzwfs[bz_index2]
            phase_c = (i2_c % size - i2_c) // size
            u2_nR = wf2.u_nR
            if phase_c.any():
                u2_nR = u2_nR * gd.plane_wave(phase_c)
            Z_kdnn[bz_index1, d] = gd.integrate(wf1.u_nR, u2_nR,
                                                global_integral=False)

            for a, P1_ni in wf1.projections.items():
                dO_ii = setups[a].dO_ii
                P2_ni = wf2.projections[a]
                Z_nn = P1_ni.conj().dot(dO_ii).dot(P2_ni.T).astype(complex)
                if phase_c.any():
                    Z_nn *= np.exp(2j * np.pi * phase_c.dot(spos_ac[a]))
                Z_kdnn[bz_index1, d] += Z_nn

        for a, P1_ni in wf1.projections.items():
            indices = proj_indices_a[a]
            m = offsets[a]
            proj_kmn[bz_index1, m:m + len(indices)] = P1_ni.T[indices]

    gd.comm.sum(Z_kdnn)
    gd.comm.sum(proj_kmn)

    overlaps = WannierOverlaps(calc.atoms,
                               nwannier,
                               kd.N_c,
                               kd.bzk_kc,
                               calc.get_fermi_level(),
                               directions,
                               Z_kdnn,
                               proj_kmn,
                               proj_indices_a)
    return overlaps


def find_directions(icell: Array2D,
                    mpsize: Sequence[int]) -> List[Tuple[int, ...]]:
    """Find nearest neighbors k-points.

    icell:
        Reciprocal cell.
    mpsize:
        Size of Monkhorst-Pack grid.

    If dk is a vector pointing at a neighbor k-points then we don't
    also include -dk in the list.  Examples: for simple cubic there
    will be 3 neighbors and for FCC there will be 6.

    For a hexagonal cell you get three directions in plane and one
    out of plane:

    >>> hex = np.array([[1, 0, 0], [0.5, 3**0.5 / 2, 0], [0, 0, 1]])
    >>> dirs = find_directions(hex, (4, 4, 4))
    >>> sorted(dirs)
    [(0, 0, 1), (0, 1, 0), (1, -1, 0), (1, 0, 0)]
    """

    from scipy.spatial import Voronoi

    d_ic = np.indices((3, 3, 3)).reshape((3, -1)).T - 1
    d_iv = d_ic.dot((icell.T / mpsize).T)
    voro = Voronoi(d_iv)
    directions: List[Tuple[int, ...]] = []
    for i1, i2 in voro.ridge_points:
        if i1 == 13 and i2 > 13:
            directions.append(tuple(d_ic[i2]))
        elif i2 == 13 and i1 > 13:
            directions.append(tuple(d_ic[i1]))
    return directions


class WaveFunction:
    def __init__(self,
                 u_nR,
                 projections: Projections):
        self.u_nR = u_nR
        self.projections = projections

    def transform(self,
                  kd: KPointDescriptor,
                  gd,
                  setups: List[Setup],
                  spos_ac: Array2D,
                  bz_index: int) -> 'WaveFunction':
        """Transforms PAW projections from IBZ to BZ k-point."""
        a_a, U_aii, time_rev = construct_symmetry_operators(
            kd, setups, spos_ac, bz_index)

        projections = self.projections.new()

        if projections.atom_partition.comm.rank == 0:
            a = 0
            for b, U_ii in zip(a_a, U_aii):
                P_msi = self.projections[b].dot(U_ii)
                if time_rev:
                    P_msi = P_msi.conj()
                projections[a][:] = P_msi
                a += 1
        else:
            assert len(projections.indices) == 0

        u_nR = 1 / 0
        return WaveFunction(u_nR, projections)

    def redistribute_atoms(self,
                           gd,
                           atom_partition: AtomPartition
                           ) -> 'WaveFunction':
        projections = self.projections.redist(atom_partition)
        u_nR = gd.distribute(self.u_nR)
        return WaveFunction(u_nR, projections)


class BZRealSpaceWaveFunctions:
    """Container for wave-functions and PAW projections (all of BZ)."""
    def __init__(self,
                 kd: KPointDescriptor,
                 gd,
                 wfs: Dict[int, WaveFunction]):
        self.kd = kd
        self.gd = gd
        self.wfs = wfs

    def __getitem__(self, bz_index):
        return self.wfs[bz_index]

    @classmethod
    def from_calculation(cls,
                         calc: GPAW,
                         n1: int = 0,
                         n2: int = 0,
                         spin=0) -> 'BZRealSpaceWaveFunctions':
        wfs = calc.wfs
        kd = wfs.kd

        if wfs.mode == 'lcao' and not wfs.positions_set:
            calc.initialize_positions()

        gd = wfs.gd.new_descriptor(comm=calc.world)

        nproj_a = wfs.kpt_qs[0][0].projections.nproj_a
        # All atoms on rank-0:
        rank_a = np.zeros_like(nproj_a)
        atom_partition = AtomPartition(gd.comm, rank_a)

        rank_a = np.arange(len(rank_a)) % gd.comm.size
        atom_partition2 = AtomPartition(gd.comm, rank_a)

        spos_ac = calc.spos_ac
        setups = calc.wfs.setups

        u_nR = gd.empty((n2 - n1), complex, global_array=True)

        bzwfs = {}
        for ibz_index in range(kd.nibzkpts):
            for n in range(n1, n2):
                u_nR[n - n1] = wfs.get_wave_function_array(n=n,
                                                           k=ibz_index,
                                                           s=spin,
                                                           periodic=True)
            P_nI = wfs.collect_projections(ibz_index, spin)
            if P_nI is not None:
                P_nI = P_nI[n1:n2]
            projections = Projections(
                nbands=n2 - n1,
                nproj_a=nproj_a,
                atom_partition=atom_partition,
                data=P_nI)

            wf = WaveFunction(u_nR.copy(), projections)

            for bz_index, ibz_index2 in enumerate(kd.bz2ibz_k):
                if ibz_index2 != ibz_index:
                    continue
                if kd.ibz2bz_k[ibz_index] == bz_index:
                    wf1 = wf
                else:
                    wf1 = wf.transform(kd, gd, setups, spos_ac, bz_index)

                bzwfs[bz_index] = wf1.redistribute_atoms(gd, atom_partition2)

        return BZRealSpaceWaveFunctions(kd, gd, bzwfs)



from ase.io import read
from gpaw import GPAW, FermiDirac
ncpus = 8
atoms = read('biimtf.xyz')
atoms.center(vacuum=5)
atoms.calc = GPAW(h=0.16,
                  charge=+1,
                  occupations=FermiDirac(0.05),
                  xc='RPBE')



from ase.build import mx2
from gpaw import GPAW, PW, FermiDirac


structure = mx2(formula='MoS2', kind='2H', a=3.184, thickness=3.127,
                size=(1, 1, 1), vacuum=5)
structure.pbc = (1, 1, 1)

ecut = 600

calc = GPAW(mode=PW(ecut),
            xc='LDA',
            kpts={'size': (30, 30, 1), 'gamma': True},
            occupations=FermiDirac(0.01),
            parallel={'domain': 1},
            txt='MoS2_gs_out.txt')

structure.calc = calc
structure.get_potential_energy()

calc.diagonalize_full_hamiltonian(nbands=500, expert=True)
calc.write('MoS2_gs_fulldiag.gpw', 'all')



import numpy as np
from ase import Atoms
from ase.units import Bohr
from gpaw.jellium import JelliumSlab
from gpaw import GPAW, PW

rs = 5.0 * Bohr  # Wigner-Seitz radius
h = 0.2          # grid-spacing
a = 8 * h        # lattice constant
v = 3 * a        # vacuum
L = 10 * a       # thickness
k = 12           # number of k-points (k*k*1)

ne = a**2 * L / (4 * np.pi / 3 * rs**3)

eps = 0.001  # displace surfaces away from grid-points
jellium = JelliumSlab(ne, z1=v - eps, z2=v + L + eps)

surf = Atoms(pbc=(True, True, False),
             cell=(a, a, v + L + v))
surf.calc = GPAW(mode=PW(400.0),
                 background_charge=jellium,
                 xc='LDA_X+LDA_C_WIGNER',
                 eigensolver='dav',
                 kpts=[k, k, 1],
                 h=h,
                 convergence={'density': 0.001},
                 nbands=int(ne / 2) + 15,
                 txt='surface.txt')
e = surf.get_potential_energy()
surf.calc.write('surface.gpw')



from ase import Atoms
from gpaw import GPAW

# Beryllium atom
atoms = Atoms(symbols='Be',
              positions=[(0, 0, 0)],
              pbc=False)

# Add 4.0 ang vacuum around the atom
atoms.center(vacuum=4.0)

# Create GPAW calculator
calc = GPAW(nbands=10, h=0.3)
# Attach calculator to atoms
atoms.calc = calc

# Calculate the ground state
energy = atoms.get_potential_energy()

# converge also the empty states (the density is converged already)
calc = calc.fixed_density(
    convergence={'bands': 8})

# Save the ground state
calc.write('Be_gs_8bands.gpw', 'all')



# Creates: graphite_EELS.png
import numpy as np
import matplotlib.pyplot as plt

plt.figure(figsize=(5, 7))
Q = np.loadtxt('graphite_q_list')
for i, q in enumerate(Q):
    filename = 'graphite_EELS_' + str(i + 1)
    d = np.loadtxt(filename, delimiter=',')
    plt.plot(d[:, 0], d[:, 2] + 0.4 * (4 - i), '-',
             label='%.2f Ang$^{-1}$' % q)

plt.xlabel('Energy [eV]')
plt.ylabel('Loss Function')
plt.title(r'EELS spectra of graphene: $\Gamma-\mathrm{M}$')
plt.legend(loc='best')
plt.savefig('graphite_EELS.png')
plt.show()



from ase.io import read
from gpaw import GPAW

atoms = read('Pt_H2O.xyz')
atoms.set_cell([[8.527708, 0, 0],
               [0, 4.923474, 0],
               [0, 0, 16]],
               scale_atoms=False)
atoms.center(axis=2)

atoms.pbc = (True, True, False)

atoms.calc = GPAW(h=0.20,
                  kpts=(2, 4, 1),
                  xc='RPBE',
                  poissonsolver={'dipolelayer': 'xy'},
                  basis='dzp',
                  maxiter=200,
                  txt='Pt_H2O.txt')
ncpus = 8



"""Test of ase wannier using gpaw."""
import numpy as np
import pytest
from ase.build import molecule
from ase.dft import Wannier

from gpaw.mpi import world
from gpaw import GPAW
from gpaw.test import equal


pytestmark = pytest.mark.skipif(world.size > 1,
                                reason='world.size > 1')


def test_ase_features_asewannier(in_tmp_dir):
    calc = GPAW(gpts=(32, 32, 32), nbands=4)
    atoms = molecule('H2', calculator=calc)
    atoms.center(vacuum=3.)
    e = atoms.get_potential_energy()

    pos = atoms.positions + np.array([[0, 0, .2339], [0, 0, -.2339]])
    com = atoms.get_center_of_mass()

    wan = Wannier(nwannier=2, calc=calc, initialwannier='bloch')
    equal(wan.get_functional_value(), 2.964, 1e-3)
    equal(np.linalg.norm(wan.get_centers() - [com, com]), 0, 1e-4)

    wan = Wannier(nwannier=2, calc=calc, initialwannier='projectors')
    equal(wan.get_functional_value(), 3.100, 2e-3)
    equal(np.linalg.norm(wan.get_centers() - pos), 0, 1e-3)

    wan = Wannier(nwannier=2,
                  calc=calc,
                  initialwannier=[[0, 0, .5], [1, 0, .5]])
    equal(wan.get_functional_value(), 3.100, 2e-3)
    equal(np.linalg.norm(wan.get_centers() - pos), 0, 1e-3)

    wan.localize()
    equal(wan.get_functional_value(), 3.100, 2e-3)
    equal(np.linalg.norm(wan.get_centers() - pos), 0, 1e-3)
    equal(np.linalg.norm(wan.get_radii() - 1.2393), 0, 2e-3)
    eig = np.sort(np.linalg.eigvals(wan.get_hamiltonian().real))
    equal(np.linalg.norm(eig - calc.get_eigenvalues()[:2]), 0, 1e-4)

    wan.write_cube(0, 'H2.cube')

    energy_tolerance = 0.002
    equal(e, -6.652, energy_tolerance)



# Creates: N2Ru_hollow.png, 2NadsRu.png, TS.xyz
from myqueue.task import task


def create_tasks():
    return [
        task('check_convergence.py', tmax='5h', cores=8),
        task('convergence.py', deps='check_convergence.py'),
        task('n2_on_metal.py', tmax='6h'),
        task('neb.py', tmax='3h', cores=8, deps='n2_on_metal.py'),
        task('vibrations.py', tmax='9h', cores=8, deps='neb.py')]



import pytest
from ase import Atoms
from gpaw import GPAW, PW
from gpaw.mpi import serial_comm
from gpaw.test import equal


@pytest.mark.skip(reason='TODO')
def test_ralda_C6_He():
    from gpaw.xc.fxc_correlation_energy import FXCCorrelation
    ecut = 50

    He = Atoms('He')
    He.center(vacuum=1.0)

    calc = GPAW(mode=PW(force_complex_dtype=True),
                xc='PBE',
                communicator=serial_comm)
    He.calc = calc
    He.get_potential_energy()
    calc.diagonalize_full_hamiltonian()

    ralda = FXCCorrelation(calc, xc='rALDA')
    C6_ralda, C6_0 = ralda.get_C6_coefficient(ecut=ecut,
                                              direction=2)

    equal(C6_0, 1.772, 0.01)
    equal(C6_ralda, 1.609, 0.01)



"""Check TDDFT ionizations with Yukawa potential."""
from ase import Atoms
from ase.units import Hartree
from gpaw import GPAW
from gpaw.mpi import world
from gpaw.cluster import Cluster
from gpaw.occupations import FermiDirac
from gpaw.test import equal
from gpaw.eigensolvers import RMMDIIS
from gpaw.lrtddft import LrTDDFT
import _gpaw


def test_rsf_yukawa_lrtddft_short(in_tmp_dir):
    libxc_version = getattr(_gpaw, 'libxc_version', '2.x.y')
    if int(libxc_version.split('.')[0]) < 3:
        from unittest import SkipTest
        raise SkipTest

    o_plus = Cluster(Atoms('Be', positions=[[0, 0, 0]]))
    o_plus.set_initial_magnetic_moments([1.0])
    o_plus.minimal_box(2.5, h=0.35)

    def get_paw():
        """Return calculator object."""
        c = {'energy': 0.05, 'eigenstates': 0.05, 'density': 0.05}
        return GPAW(convergence=c, eigensolver=RMMDIIS(),
                    nbands=3,
                    xc='PBE',
                    #                experimental={'niter_fixdensity': 2},
                    parallel={'domain': world.size}, h=0.35,
                    occupations=FermiDirac(width=0.0, fixmagmom=True))

    calc_plus = get_paw()
    calc_plus.set(txt='Be_plus_LCY_PBE_083.log', charge=1)
    o_plus.calc = calc_plus
    o_plus.get_potential_energy()
    calc_plus.set(xc='LCY-PBE:omega=0.83:unocc=True',
                  experimental={'niter_fixdensity': 2})
    o_plus.get_potential_energy()
    lr = LrTDDFT(calc_plus, txt='LCY_TDDFT_Be.log',
                 restrict={'istart': 0, 'jend': 1})
    equal(lr.xc.omega, 0.83)
    lr.write('LCY_TDDFT_Be.ex.gz')
    e_ion = 9.3
    ip_i = 13.36
    # reading is problematic with EXX on more than one core
    if world.rank == 0:
        lr2 = LrTDDFT.read('LCY_TDDFT_Be.ex.gz')
        lr2.diagonalize()
        equal(lr2.xc.omega, 0.83)
        ion_i = lr2[0].get_energy() * Hartree + e_ion
        equal(ion_i, ip_i, 0.3)



"""GBAR customize.py for GPAW."""

from pathlib import Path

xc = str(Path.home()) + '/'  # libxc folder
libraries = ['xc', 'openblas', 'scalapack']
library_dirs = [
    xc + 'lib',
    '/appl/ScaLAPACK/2.0.2-openblas-0.2.15//XeonE5-2670v3/gcc-6.1.0/lib',
    '/appl/OpenBLAS/0.2.14/XeonE5-2670v3/gcc/lib']
scalapack = True
define_macros += [('GPAW_NO_UNDERSCORE_CBLACS', '1')]
define_macros += [('GPAW_NO_UNDERSCORE_CSCALAPACK', '1')]
include_dirs += [xc + 'include']
library_dirs += [xc + 'lib']
extra_link_args += [f'-Wl,-rpath={xc}/lib']



from gpaw.response.g0w0 import G0W0

gw = G0W0(calc='C_groundstate.gpw',
          nbands=30,      # number of bands for calculation of self-energy
          bands=(3, 5),   # VB and CB
          ecut=20.0,      # plane-wave cutoff for self-energy
          filename='C-g0w0',
          savepckl=True)  # save a .pckl file with results
result = gw.calculate()



# Creates: bs_si.png
"""GLLBSC band gap tutorial

Calculate the band structure and band gap of Si using GLLBSC.
"""
# P1
from ase.build import bulk
from gpaw import GPAW, PW, FermiDirac

# Ground state calculation
atoms = bulk('Si', 'diamond', 5.431)
calc = GPAW(mode=PW(200),
            xc='GLLBSC',
            kpts={'size': (8, 8, 8), 'gamma': True},
            occupations=FermiDirac(0.01),
            txt='gs.out')
atoms.calc = calc
atoms.get_potential_energy()

# Band structure calculation with fixed density
bs_calc = calc.fixed_density(nbands=10,
                             kpts={'path': 'LGXWKL', 'npoints': 60},
                             symmetry='off',
                             convergence={'bands': 8},
                             txt='bs.out')

# Plot the band structure
bs = bs_calc.band_structure().subtract_reference()
bs.plot(filename='bs_si.png', emin=-6, emax=6)
# P2
# Get the accurate HOMO and LUMO from the band structure calculator
homo, lumo = bs_calc.get_homo_lumo()

# Calculate the discontinuity potential using the ground state calculator and
# the accurate HOMO and LUMO
response = calc.hamiltonian.xc.response
dxc_pot = response.calculate_discontinuity_potential(homo, lumo)

# Calculate the discontinuity using the band structure calculator
bs_response = bs_calc.hamiltonian.xc.response
KS_gap, dxc = bs_response.calculate_discontinuity(dxc_pot)

# Fundamental band gap = Kohn-Sham band gap + derivative discontinuity
QP_gap = KS_gap + dxc

print(f'Kohn-Sham band gap:         {KS_gap:.2f} eV')
print(f'Discontinuity from GLLB-sc: {dxc:.2f} eV')
print(f'Fundamental band gap:       {QP_gap:.2f} eV')



#User provided customizations for the gpaw setup

#Here, one can override the default arguments, or append own
#arguments to default ones
#To override use the form
#     libraries = ['somelib','otherlib']
#To append use the form
#     libraries += ['somelib','otherlib']

compiler = 'mpicc'
libraries = ['mkl_intel_lp64' ,'mkl_sequential' ,'mkl_core',
             'mkl_lapack',
             'mkl_scalapack_lp64', 'mkl_blacs_intelmpi_lp64',
             'pthread', 
             'readline', 'termcap',
             'xc'
             ]
#libraries = []
#libraries += []

#library_dirs = []
#library_dirs += []

#include_dirs = []
#include_dirs += []

#extra_link_args = []
#extra_link_args += []

#extra_compile_args = []
#extra_compile_args += []

#runtime_library_dirs = []
#runtime_library_dirs += []

#extra_objects = []
#extra_objects += []

#define_macros = []
#define_macros += []

#mpicompiler = None
#mpilinker = None
#mpi_libraries = []
#mpi_libraries += []

#mpi_library_dirs = []
#mpi_library_dirs += []

#mpi_include_dirs = []
#mpi_include_dirs += []

#mpi_runtime_library_dirs = []
#mpi_runtime_library_dirs += []

#mpi_define_macros = []
#mpi_define_macros += []

#platform_id = ''

#hdf5 = True

# Valid values for scalapack are False, or True:
# False (the default) - no ScaLapack compiled in
# True - ScaLapack compiled in
# Warning! At least scalapack 2.0.1 is required!
# See https://trac.fysik.dtu.dk/projects/gpaw/ticket/230
scalapack = False

if scalapack:
    libraries += ['scalapack']
    library_dirs += []
    define_macros += [('GPAW_NO_UNDERSCORE_CBLACS', '1')]
    define_macros += [('GPAW_NO_UNDERSCORE_CSCALAPACK', '1')]

# In order to link libxc installed in a non-standard location
# (e.g.: configure --prefix=/home/user/libxc-2.0.1-1), use:
# - static linking:
#include_dirs += ['/home/user/libxc-2.0.1-1/include']
#extra_link_args += ['/home/user/libxc-2.0.1-1/lib/libxc.a']
#if 'xc' in libraries: libraries.remove('xc')
# - dynamic linking (requires also setting LD_LIBRARY_PATH at runtime):
#include_dirs += ['/home/user/libxc-2.0.1-1/include']
#library_dirs += ['/home/user/libxc-2.0.1-1/lib']
#if 'xc' not in libraries: libraries.append('xc')



from ase import Atoms
from ase.parallel import paropen
from gpaw import GPAW, FermiDirac
from gpaw.mixer import MixerSum
from gpaw.hybrids.energy import non_self_consistent_energy as nsc_energy
from gpaw.wavefunctions.pw import PW

# CO

CO = Atoms('CO', [(0, 0, 0), (0, 0, 1.1283)])
CO.set_pbc(True)
CO.center(vacuum=3.0)
calc = GPAW(mode=PW(600, force_complex_dtype=True),
            parallel={'domain': 1},
            xc='PBE',
            txt='CO.ralda_01_CO_pbe.txt',
            convergence={'density': 1.e-6})

CO.calc = calc
E0_pbe = CO.get_potential_energy()

E0_hf = nsc_energy(calc, 'EXX')

calc.diagonalize_full_hamiltonian()
calc.write('CO.ralda.pbe_wfcs_CO.gpw', mode='all')

# C

C = Atoms('C')
C.set_pbc(True)
C.set_cell(CO.cell)
C.center()
calc = GPAW(mode=PW(600, force_complex_dtype=True),
            parallel={'domain': 1},
            xc='PBE',
            mixer=MixerSum(beta=0.1, nmaxold=5, weight=50.0),
            hund=True,
            occupations=FermiDirac(0.01, fixmagmom=True),
            txt='CO.ralda_01_C_pbe.txt',
            convergence={'density': 1.e-6})

C.calc = calc
E1_pbe = C.get_potential_energy()

E1_hf = nsc_energy(calc, 'EXX')

f = paropen('CO.ralda.PBE_HF_C.dat', 'w')
print(E1_pbe, E1_hf, file=f)
f.close()

calc.diagonalize_full_hamiltonian()
calc.write('CO.ralda.pbe_wfcs_C.gpw', mode='all')

# O

O = Atoms('O')
O.set_pbc(True)
O.set_cell(CO.cell)
O.center()
calc = GPAW(mode=PW(600, force_complex_dtype=True),
            parallel={'domain': 1},
            xc='PBE',
            mixer=MixerSum(beta=0.1, nmaxold=5, weight=50.0),
            hund=True,
            txt='CO.ralda_01_O_pbe.txt',
            convergence={'density': 1.e-6})

O.calc = calc
E2_pbe = O.get_potential_energy()

E2_hf = nsc_energy(calc, 'EXX')

calc.diagonalize_full_hamiltonian()
calc.write('CO.ralda.pbe_wfcs_O.gpw', mode='all')

f = paropen('CO.ralda.PBE_HF_CO.dat', 'w')
print('PBE: ', E0_pbe - E1_pbe - E2_pbe, file=f)
print('HF: ', E0_hf - E1_hf - E2_hf, file=f)
f.close()



# Creates: ind_1.12.png, ind_2.48.png
# -*- coding: utf-8 -*-
import numpy as np
import matplotlib.pyplot as plt

from ase.io import read


def do(freq):
    # Read cube file
    cube = read(f'ind_{freq:.2f}.cube', full_output=True)
    d_g = cube['data']
    atoms = cube['atoms']
    box = np.diag(atoms.get_cell())
    ng = d_g.shape

    # Take slice of data array
    d_yx = d_g[:, :, ng[2] // 2]
    x = np.linspace(0, box[0], ng[0])
    xlabel = u'x (Å)'
    y = np.linspace(0, box[1], ng[1])
    ylabel = u'y (Å)'

    # Plot
    plt.figure(figsize=(8, 3.5))
    ax = plt.subplot(1, 1, 1)
    X, Y = np.meshgrid(x, y)
    dmax = max(d_yx.min(), d_yx.max())
    vmax = 0.9 * dmax
    vmin = -vmax
    plt.pcolormesh(X, Y, d_yx.T, cmap='RdBu_r', vmin=vmin, vmax=vmax)
    contours = np.sort(np.outer([-1, 1], [0.02]).ravel() * dmax)
    plt.contour(X, Y, d_yx.T, contours, cmap='RdBu_r', vmin=-1e-10, vmax=1e-10)
    for atom in atoms:
        pos = atom.position
        plt.scatter(pos[0], pos[1], s=100, c='k', marker='o')
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.xlim([x[0], x[-1]])
    plt.ylim([y[0], y[-1]])
    ax.set_aspect('equal')

    plt.title(f'Induced density of Na8 at {freq:.2f} eV')
    plt.tight_layout()
    plt.savefig(f'ind_{freq:.2f}.png')


do(1.12)
do(2.48)



from ase.build import bulk
from ase.optimize import BFGS
from gpaw import GPAW, PW


def nv_minus(n: int, relax: bool = False) -> None:
    atoms = bulk('C', 'diamond', cubic=True) * n
    atoms.numbers[0] = 7
    del atoms[1]
    atoms.set_initial_magnetic_moments([1] * len(atoms))
    name = f'NC{8 * n**3 - 2}'
    atoms.calc = GPAW(xc='PBE',
                      mode=PW(ecut=400),
                      charge=-1,
                      txt=name + '.txt')
    atoms.get_forces()
    atoms.calc.write(name + '.gpw', 'all')
    if relax:
        BFGS(atoms,
             logfile=name + '.log').run(fmax=0.05)
        atoms.calc.write(name + '.relaxed.gpw', 'all')


if __name__ == '__main__':
    for n in range(2, 4):
        nv_minus(n, relax=True)



import numpy as np

from ase.build import molecule
from gpaw import GPAW
from gpaw.lcaotddft import LCAOTDDFT
from gpaw.poisson import PoissonSolver
from gpaw.lcaotddft.dipolemomentwriter import DipoleMomentWriter
from gpaw.mpi import world

from gpaw.test import equal

# Atoms


def test_lcaotddft_fxc_is_xc(in_tmp_dir):
    atoms = molecule('Na2')
    atoms.center(vacuum=4.0)

    # Ground-state calculation
    calc = GPAW(nbands=2, h=0.4, setups=dict(Na='1'),
                basis='dzp', mode='lcao',
                poissonsolver=PoissonSolver(eps=1e-16),
                convergence={'density': 1e-8},
                xc='LDA',
                txt='gs.out')
    atoms.calc = calc
    atoms.get_potential_energy()
    calc.write('gs.gpw', mode='all')

    # Time-propagation calculation without fxc
    td_calc = LCAOTDDFT('gs.gpw', txt='td.out')
    DipoleMomentWriter(td_calc, 'dm.dat')
    td_calc.absorption_kick(np.ones(3) * 1e-5)
    td_calc.propagate(20, 4)
    world.barrier()

    # Check dipole moment file
    ref = np.loadtxt('dm.dat').ravel()

    # Time-propagation calculation with fxc=xc
    td_calc = LCAOTDDFT('gs.gpw', fxc='LDA', txt='td_fxc.out')
    DipoleMomentWriter(td_calc, 'dm_fxc.dat')
    td_calc.absorption_kick(np.ones(3) * 1e-5)
    td_calc.propagate(20, 1)
    td_calc.write('td_fxc.gpw', mode='all')

    # Restart from the restart point
    td_calc = LCAOTDDFT('td_fxc.gpw', txt='td_fxc2.out')
    DipoleMomentWriter(td_calc, 'dm_fxc.dat')
    td_calc.propagate(20, 3)

    # Check dipole moment file
    data = np.loadtxt('dm_fxc.dat')[[0, 1, 2, 4, 5, 6]].ravel()

    tol = 1e-9
    equal(data, ref, tol)



import os
import shutil
from pathlib import Path
from myqueue.task import task


def create_tasks():
    if os.getenv('AGTS_FILES'):
        dir = Path(os.getenv('AGTS_FILES'))
        file = Path('organometal.db')
        if not file.is_file():
            shutil.copyfile(dir / file, file)
    return [task('machinelearning.py', tmax='8h')]



compiler='mpicc'
libraries = ['xc']
extra_link_args += [
     '-Wl,--no-as-needed',
    '-lmkl_scalapack_lp64',
    '-lmkl_intel_lp64',
    '-lmkl_core',
    '-lmkl_sequential',
    '-lmkl_blacs_intelmpi_lp64',
]



# Creates: Ag55_spec.png
# -*- coding: utf-8 -*-
import numpy as np
import matplotlib.pyplot as plt

data_ej = np.loadtxt('spec.dat')

plt.figure(figsize=(6, 6 / 1.62))
ax = plt.subplot(1, 1, 1)
ax.plot(data_ej[:, 0], data_ej[:, 1], 'k')
ax.spines['right'].set_visible(False)
ax.spines['top'].set_visible(False)
ax.yaxis.set_ticks_position('left')
ax.xaxis.set_ticks_position('bottom')
plt.title(r'Absorption spectrum of Ag$_{55}$ with GLLB-SC potential')
plt.xlabel('Energy (eV)')
plt.ylabel('Photoabsorption (eV$^{-1}$)')
plt.xlim(0, 6)
plt.ylim(ymin=0)
plt.tight_layout()
plt.savefig('Ag55_spec.png')



from gpaw.lrtddft import LrTDDFT
from gpaw.lrtddft import photoabsorption_spectrum

lr = LrTDDFT.read('lr.dat.gz')
lr.diagonalize()

# write the spectrum to the data file
photoabsorption_spectrum(lr,
                         'spectrum_w.05eV.dat',  # data file name
                         width=0.05)             # width in eV



from gpaw.xc.fxc import FXCCorrelation

fxc = FXCCorrelation('diam_kern.ralda.lda_wfcs.gpw', xc='rALDA',
                     txt='diam_kern.ralda_02_ralda_dens.txt')
E_i = fxc.calculate(ecut=[131.072])



from ase.units import Hartree
from gpaw import GPAW
from gpaw.utilities.dos import fold
import pickle
import matplotlib.pyplot as plt

e_f = GPAW('top.gpw').get_fermi_level()

e_n, P_n = pickle.load(open('top.pickle', 'rb'))
for n in range(2, 7):
    e, ldos = fold(e_n[n] * Hartree, P_n[n], npts=2001, width=0.2)
    plt.plot(e - e_f, ldos, label='Band: ' + str(n))
plt.legend()
plt.axis([-15, 10, None, None])
plt.xlabel('Energy [eV]')
plt.ylabel('PDOS')
plt.show()



from ase import Atoms
from ase.optimize import QuasiNewton
from gpaw import GPAW

a = 6
b = a / 2

mol = Atoms('H2O',
            [(b, 0.7633 + b, -0.4876 + b),
             (b, -0.7633 + b, -0.4876 + b),
             (b, b, 0.1219 + b)],
            cell=[a, a, a])

calc = GPAW(nbands=4,
            h=0.2,
            mode='lcao',
            basis='dzp')

mol.calc = calc
dyn = QuasiNewton(mol, trajectory='lcao_h2o.traj')
dyn.run(fmax=0.05)



import pytest
import numpy as np
from ase.build import molecule
from ase.parallel import parprint, world

from gpaw import GPAW
from gpaw.test import equal
from gpaw.cluster import Cluster
from gpaw.analyse.overlap import Overlap
from gpaw.lrtddft import LrTDDFT


@pytest.mark.skip(reason='TODO')
def test_overlap():
    """Evaluate the overlap between two independent calculations

    Differences are forced by different eigensolvers and differing number
    of Kohn-Sham states.
    """

    h = 0.4
    box = 2
    nbands = 4
    txt = '-'
    txt = None
    np.set_printoptions(precision=3, suppress=True)

    H2 = Cluster(molecule('H2'))
    H2.minimal_box(box, h)

    c1 = GPAW(h=h, txt=txt, eigensolver='dav', nbands=nbands,
              convergence={'eigenstates': nbands})
    c1.calculate(H2)
    fname = 'H2.gpw'
    c1.write(fname, 'all')
    lr1 = LrTDDFT(c1)

    parprint('sanity --------')
    ov = Overlap(c1).pseudo(c1)
    parprint('pseudo(normalized):\n', ov)
    ov = Overlap(c1).pseudo(c1, False)
    parprint('pseudo(not normalized):\n', ov)
    ov = Overlap(c1).full(c1)
    parprint('full:\n', ov)
    equal(ov[0], np.eye(ov[0].shape[0], dtype=ov.dtype), 1e-10)

    parprint('io --------')
    c1copy = GPAW(fname, txt=txt)
    ov = Overlap(c1).full(c1copy)
    parprint('full:\n', ov)
    """require the overlap matrix of yourself and your read copy
    to be the unity matrix"""
    equal(ov[0], np.eye(ov[0].shape[0], dtype=ov.dtype), 1e-10)

    def show(c2):
        c2.calculate(H2)
        ov = Overlap(c1).pseudo(c2)
        parprint('wave function overlap (pseudo):\n', ov)
        ov = Overlap(c1).full(c2)
        parprint('wave function overlap (full):\n', ov)
        lr2 = LrTDDFT(c2)
        ovkss = lr1.kss.overlap(ov[0], lr2.kss)
        parprint('KSSingles overlap:\n', ovkss)
        ovlr = lr1.overlap(ov[0], lr2)
        parprint('LrTDDFT overlap:\n', ovlr)

    parprint('cg --------')
    c2 = GPAW(h=h, txt=txt, eigensolver='cg', nbands=nbands + 1,
              convergence={'eigenstates': nbands + 1})
    show(c2)

    parprint('spin --------')
    H2.set_initial_magnetic_moments([1, -1])
    c2 = GPAW(h=h, txt=txt, spinpol=True, nbands=nbands + 1,
              parallel={'domain': world.size},
              convergence={'eigenstates': nbands + 1})
    H2.set_initial_magnetic_moments([0, 0])
    try:
        show(c2)
    except AssertionError:
        parprint('Not ready')
    if 1:
        ov = Overlap(c1).pseudo(c2, otherspin=1)
        parprint('wave function overlap (pseudo other spin):\n', ov)

    parprint('k-points --------')

    H2.set_pbc([1, 1, 1])
    c1 = GPAW(h=h, txt=txt, nbands=nbands,
              kpts=(1, 1, 3),
              # parallel={'domain': world.size},
              convergence={'eigenstates': nbands})
    c1.calculate(H2)
    c2 = GPAW(h=h, txt=txt, nbands=nbands + 1,
              kpts=(1, 1, 3),
              # parallel={'domain': world.size},
              convergence={'eigenstates': nbands + 1})
    try:
        show(c2)
    except (AssertionError, IndexError) as e:
        parprint('Not ready', e)



#!/usr/bin/env python
import sys
import xml.sax
from optparse import OptionParser
from math import exp


class Reader(xml.sax.handler.ContentHandler):
    def __init__(self, list, name, state):
        self.list = list
        self.name = name
        self.state = state
        self.grid = {}
        xml.sax.handler.ContentHandler.__init__(self)

    def read(self, filename):
        if filename.endswith('.gz'):
            import gzip
            source = gzip.open(filename)
        else:
            source = open(filename)
        xml.sax.parse(source, self)

    def startElement(self, name, attrs):
        if name == 'state' and self.list:
            print(attrs['id'], file=sys.stderr)
        elif name == 'radial_grid':
            istart = int(attrs['istart'])
            iend = int(attrs['iend'])
            ii = range(istart, iend + 1)
            id = attrs['id']
            eq = attrs['eq']
            if eq == 'r=a*exp(d*i)':
                a = float(attrs['a'])
                d = float(attrs['d'])
                self.grid[id] = [a * exp(d * i) for i in ii]
            elif eq == 'r=a*i/(n-i)':
                a = float(attrs['a'])
                n = float(attrs['n'])
                self.grid[id] = [a * i / (n - i) for i in ii]
            elif eq == 'r=a*(exp(d*i)-1)':
                a = float(attrs['a'])
                d = float(attrs['d'])
                self.grid[id] = [a * (exp(d * i) - 1.0) for i in ii]
            elif eq == 'r=d*i':
                d = float(attrs['d'])
                self.grid[id] = [d * i for i in ii]
            elif eq == 'r=(i/n+a)^5/a-a^4':
                a = float(attrs['a'])
                n = float(attrs['n'])
                self.grid[id] = [(i / n + a)**5 / a - a**4 for i in ii]
            else:
                raise RuntimeError('Unknown grid type: ' + attrs['eq'])
        elif name == self.name and attrs.get('state', None) == self.state:
            self.r = self.grid[attrs['grid']]
            self.data = ''
        else:
            self.data = None

    def characters(self, data):
        if self.data is not None:
            self.data += data

    def endElement(self, name):
        if self.data is not None:
            for r, x in zip(self.r, self.data.split()):
                print(r, x)


op = OptionParser(usage='%prog [options] setup[.gz]',
                  version='%prog 0.2')
op.add_option('-x', '--extract',
              help='Function to extract.',
              metavar='<name>')
op.add_option('-s', '--state',
              help='Select valence state.',
              metavar='<channel>')
op.add_option('-l', '--list', action='store_true',
              help='List valence states.')
options, args = op.parse_args()
if len(args) != 1:
    op.error('incorrect number of arguments')


reader = Reader(options.list, options.extract, options.state)
reader.read(args[0])



import pytest
from gpaw.mpi import world
from gpaw.utilities import compiled_with_sl
from ase import Atoms
from ase.build import molecule
from gpaw import GPAW, PW
from gpaw.xc.fxc import FXCCorrelation
from gpaw.test import equal

pytestmark = pytest.mark.skipif(
    world.size != 1 and not compiled_with_sl(),
    reason='world.size != 1 and not compiled_with_sl()')


def test_ralda_ralda_energy_N2(in_tmp_dir):
    if world.size == 1:
        scalapack1 = None
        scalapack2 = None
    elif world.size == 2:
        scalapack1 = (2, world.size // 2, 32)
        scalapack2 = None
    else:
        scalapack1 = (2, world.size // 2, 32)
        scalapack2 = (2, world.size // 2, 32)

    # N2
    N2 = molecule('N2')
    N2.set_cell((2.5, 2.5, 3.5))
    N2.center()
    calc = GPAW(mode=PW(force_complex_dtype=True),
                eigensolver='rmm-diis',
                xc='LDA',
                nbands=16,
                basis='dzp',
                parallel={'domain': 1},
                convergence={'density': 1.e-6})
    N2.calc = calc
    N2.get_potential_energy()
    calc.diagonalize_full_hamiltonian(nbands=80, scalapack=scalapack1)
    calc.write('N2.gpw', mode='all')

    ralda = FXCCorrelation('N2.gpw', xc='rALDA')
    Ec_N2 = ralda.calculate(ecut=[50])

    # N
    N = Atoms('N', [(0, 0, 0)])
    N.set_cell((2.5, 2.5, 3.5))
    N.center()
    calc = GPAW(mode=PW(force_complex_dtype=True),
                eigensolver='rmm-diis',
                xc='LDA',
                basis='dzp',
                nbands=8,
                hund=True,
                parallel={'domain': 1},
                convergence={'density': 1.e-6})
    N.calc = calc
    N.get_potential_energy()
    calc.diagonalize_full_hamiltonian(nbands=80, scalapack=scalapack2)
    calc.write('N.gpw', mode='all')

    ralda = FXCCorrelation('N.gpw', xc='rALDA')
    Ec_N = ralda.calculate(ecut=[50])

    equal(Ec_N2, -6.17, 0.01,)
    equal(Ec_N, -1.1085, 0.001)



#!/usr/bin/python
"""icc.py is a wrapper for the Intel compiler,
   converting/removing incompatible gcc args.   """

import sys
from subprocess import call

args2change = {"-fno-strict-aliasing":"",
               "-fmessage-length=0":"",
               "-Wall":"",
               "-std=c99":"-qlanglvl=extc99",
               "-fPIC":"",
               "-g":"",
               "-D_FORTIFY_SOURCE=2":"",
               "-DNDEBUG":"",
               "-UNDEBUG":"",
               "-pthread":"",
               "-shared":"-qmkshrobj",
               "-Xlinker":"",
               "-export-dynamic":"",
               "-Wstrict-prototypes":"",
               "-dynamic":"",
               "-O3":"",
               "-O3":"",
               "-O2":"",
               "-O1":""}

fragile_files = ["test.c"]

cmd = ""
fragile = False
for arg in sys.argv[1:]:
    cmd += " "
    t = arg.strip()
    if t in fragile_files:
        fragile = True
    if t in args2change:
        cmd += args2change[t]
    else:
        cmd += arg

flags = "-w -O3 -std=c99"
cmd = f"mpicc {flags} {cmd}"

call(cmd, shell=True)




import numpy as np

from ase.io import write
from gpaw import GPAW
from gpaw.tddft.units import au_to_eV
from gpaw.lcaotddft.densitymatrix import DensityMatrix
from gpaw.lcaotddft.frequencydensitymatrix import FrequencyDensityMatrix

# Load the objects
calc = GPAW('unocc.gpw', txt=None)
calc.initialize_positions()  # Initialize in order to calculate density
dmat = DensityMatrix(calc)
fdm = FrequencyDensityMatrix(calc, dmat, 'fdm.ulm')


def do(w):
    # Select the frequency and the density matrix
    rho_MM = fdm.FReDrho_wuMM[w][0]
    freq = fdm.freq_w[w]
    frequency = freq.freq * au_to_eV
    print(f'Frequency: {frequency:.2f} eV')
    print(f'Folding: {freq.folding}')

    # Induced density
    rho_g = dmat.get_density([rho_MM.imag])

    # Save as a cube file
    write(f'ind_{frequency:.2f}.cube', calc.atoms, data=rho_g)

    # Calculate dipole moment for reference
    dm_v = dmat.density.finegd.calculate_dipole_moment(rho_g, center=True)
    absorption = 2 * freq.freq / np.pi * dm_v[0] / au_to_eV * 1e5
    print(f'Total absorption: {absorption:.2f} eV^-1')


do(0)
do(1)



import numpy as np

from ase.build import bulk
from ase.optimize import BFGS
from ase.io import Trajectory
from ase.constraints import StrainFilter

from gpaw import GPAW, PW


co = bulk('Co')
co.set_initial_magnetic_moments([1.6, 1.6])
co.calc = GPAW(mode=PW(700),
               xc='PBE',
               kpts=(8, 8, 4),
               txt='co.txt')

BFGS(StrainFilter(co)).run(0.005)

a0 = co.cell[0, 0]
c0 = co.cell[2, 2]

traj = Trajectory('co.traj', 'w')
eps = 0.01
for a in a0 * np.linspace(1 - eps, 1 + eps, 3):
    for c in c0 * np.linspace(1 - eps, 1 + eps, 3):
        co.set_cell(bulk('Co', a=a, covera=c / a).cell, scale_atoms=True)
        co.get_potential_energy()
        traj.write(co)



import sys
import numpy as np
from ase import Atoms
from ase.build import niggli_reduce
from gpaw import GPAW, FermiDirac

# Script to get the total energies of a supercell
# of BN with and without a carbon substitution

c = 15.0

N = int(sys.argv[1])  # NxNx1 supercell
h = 0.15

a = 2.51026699
cell = [[a, 0., 0.],
        [-a / 2, np.sqrt(3) / 2 * a, 0.],
        [0., 0., c]]
scaled_positions = [[2. / 3, 1. / 3, 0.5],
                    [1. / 3, 2. / 3, 0.5]]

system = Atoms('BN',
               cell=cell,
               scaled_positions=scaled_positions,
               pbc=[1, 1, 1])
system = system.repeat((2, 1, 1))
niggli_reduce(system)

defect = system.repeat((N, N, 1))
defect[0].symbol = 'C'
defect[1].magmom = 1

q = +1  # Defect charge

calc = GPAW(mode='fd',
            kpts={'size': (4, 4, 1)},
            xc='PBE',
            charge=q,
            occupations=FermiDirac(0.01),
            txt='BN{0}{0}.CB_charged.txt'.format(N))


defect.calc = calc
defect.get_potential_energy()
calc.write('BN{0}{0}.C_B_charged.gpw'.format(N))

# Neutral case
parameters = calc.todict()
parameters['txt'] = 'BN{0}{0}.CB_neutral.txt'.format(N)
parameters['charge'] = 0
calc = GPAW(**parameters)

defect.calc = calc
defect.get_potential_energy()
calc.write('BN{0}{0}.C_B_neutral.gpw'.format(N))


# Now for the pristine case

pristine = system.repeat((N, N, 1))
parameters['txt'] = 'BN{0}{0}.pristine.txt'.format(N)
calc = GPAW(**parameters)

pristine.calc = calc
pristine.get_potential_energy()
calc.write('BN{0}{0}.pristine.gpw'.format(N))



"""Check TDDFT ionizations with Yukawa potential."""
from ase.structure import molecule
from ase.units import Hartree
from gpaw import GPAW
from gpaw.mpi import world
from gpaw.cluster import Cluster
from gpaw.occupations import FermiDirac
from gpaw.test import equal
from gpaw.eigensolvers import RMMDIIS
from gpaw.lrtddft import LrTDDFT

h2o = Cluster(molecule('H2O'))
h2o.set_initial_magnetic_moments([2, -1, -1])
h2o.minimal_box(3.0, h=0.3)
h2o_plus = Cluster(molecule('H2O'))
h2o_plus.set_initial_magnetic_moments([2, -0.5, -0.5])
h2o_plus.minimal_box(3.0, h=0.3)


def get_paw():
    """Return calculator object."""
    c = {'energy': 0.001, 'eigenstates': 0.001, 'density': 0.001}
    return GPAW(convergence=c, eigensolver=RMMDIIS(),
                xc='LCY-PBE:omega=0.83:unocc=True',
                parallel={'domain': world.size}, h=0.3,
                occupations=FermiDirac(width=0.0, fixmagmom=True))


calc = get_paw()
calc.set(txt='H2O_LCY_PBE_083.log')
calc_plus = get_paw()
calc_plus.set(txt='H2O_plus_LCY_PBE_083.log', charge=1)

h2o.calc = calc
e_h2o = h2o.get_potential_energy()
h2o_plus.calc = calc_plus
e_h2o_plus = h2o_plus.get_potential_energy()
e_ion = e_h2o_plus - e_h2o

print(e_ion, 12.62)
equal(e_ion, 12.62, 0.1)
lr = LrTDDFT(calc_plus, txt='LCY_TDDFT_H2O.log', jend=4)
equal(lr.xc.omega, 0.83)
lr.write('LCY_TDDFT_H2O.ex.gz')
# reading is problematic with EXX on more than one core
if world.rank == 0:
    lr2 = LrTDDFT('LCY_TDDFT_H2O.ex.gz')
    lr2.diagonalize()
    equal(lr2.xc.omega, 0.83)

    for i, ip_i in enumerate([14.74, 18.51]):
        ion_i = lr2[i].get_energy() * Hartree + e_ion
        print(ion_i, ip_i)
        equal(ion_i, ip_i, 0.6)



# Creates: bse_MoS2.png
import matplotlib.pyplot as plt
import numpy as np

plt.figure()

a = np.loadtxt('pol_rpa_MoS2.csv', delimiter=',')
plt.plot(a[:, 0], a[:, 4], label='RPA', lw=2)

a = np.loadtxt('pol_bse_MoS2.csv', delimiter=',')
plt.plot(a[:, 0], a[:, 2], label='BSE', lw=2)

a = np.loadtxt('pol_bse_MoS2_trun.csv', delimiter=',')
plt.plot(a[:, 0], a[:, 2], label='BSE - Truncated', lw=2)

plt.xlabel(r'$\hbar\omega\;[eV]$', size=24)
plt.ylabel(r'$\alpha_2$', size=24)
plt.xticks(size=20)
plt.yticks(size=20)
plt.tight_layout()
# plt.axis([2.0, 6.0, None, None])
plt.legend(loc='upper left')

# plt.show()
plt.savefig('bse_MoS2.png')



from myqueue.task import task


def create_tasks():
    gs = task('gs.py@48:30m')
    td = task('td.py@48:4h', deps=[gs])
    spec = task('spec.py@1:1m', deps=[td])
    plot = task('plot_spec.py@1:1m', deps=[spec])
    basis_my = task('basis.py@1:10m', folder='mybasis')
    gs_my = task('gs.py@48:30m', folder='mybasis', deps=[basis_my])
    td_my = task('td.py@48:4h', folder='mybasis', deps=[gs_my])
    spec_my = task('spec.py@1:1m', folder='mybasis', deps=[td_my])
    gs_dzp = task('gs.py@48:2h', folder='dzp')
    td_dzp = task('td.py@48:4h', folder='dzp', deps=[gs_dzp])
    spec_dzp = task('spec.py@1:1m', folder='dzp', deps=[td_dzp])
    basis_plot = task('plot_spec_basis.py@1:1m',
                      deps=[spec, spec_my, spec_dzp])

    return [gs, td, spec, plot,
            basis_my, gs_my, td_my, spec_my,
            gs_dzp, td_dzp, spec_dzp,
            basis_plot]



from myqueue.task import task


def create_tasks():
    return [
        task('gs_N2.py@16:10m'),
        task('frequency.py@1:3h', deps='gs_N2.py'),
        task('con_freq.py@2:16h', deps='gs_N2.py'),
        task('rpa_N2.py@32:20h', deps='gs_N2.py'),
        task('plot_w.py', deps='frequency.py,con_freq.py'),
        task('plot_con_freq.py', deps='con_freq.py'),
        task('extrapolate.py', deps='rpa_N2.py'),
        task('shell:rm N.gpw N2.gpw',  # clean up
             deps='frequency.py,con_freq.py,rpa_N2.py')]



import time

import pytest
import numpy as np
from ase.build import bulk
from ase.parallel import parprint

from gpaw import GPAW, PW
from gpaw.test import findpeak
from gpaw.response.df import DielectricFunction
from gpaw.mpi import size, world


@pytest.mark.libxc
def test_response_aluminum_EELS_ALDA(in_tmp_dir):
    assert size <= 4**3

    # Ground state calculation

    t1 = time.time()

    a = 4.043
    atoms = bulk('Al', 'fcc', a=a)
    atoms.center()
    calc = GPAW(mode=PW(200),
                nbands=4,
                kpts=(4, 4, 4),
                parallel={'band': 1},
                idiotproof=False,  # allow uneven distribution of k-points
                xc='LDA')

    atoms.calc = calc
    atoms.get_potential_energy()
    calc.write('Al', 'all')
    t2 = time.time()

    # Excited state calculation
    q = np.array([1 / 4, 0, 0])
    w = np.linspace(0, 24, 241)

    df = DielectricFunction(calc='Al', frequencies=w, eta=0.2, ecut=50,
                            hilbert=False)
    df.get_eels_spectrum(xc='ALDA', filename='EELS_Al_ALDA.csv', q_c=q)
    # df.check_sum_rule()
    # df.write('Al.pckl')

    t3 = time.time()

    parprint('For ground  state calc, it took', (t2 - t1) / 60, 'minutes')
    parprint('For excited state calc, it took', (t3 - t2) / 60, 'minutes')

    world.barrier()
    d = np.loadtxt('EELS_Al_ALDA.csv', delimiter=',')
    # New results are compared with test values
    wpeak1, Ipeak1 = findpeak(d[:, 0], d[:, 1])
    wpeak2, Ipeak2 = findpeak(d[:, 0], d[:, 2])

    test_wpeak1 = 15.7002862696  # eV
    test_Ipeak1 = 28.5590363176
    test_wpeak2 = 15.5196459206
    test_Ipeak2 = 26.5680624522

    if abs(test_wpeak1 - wpeak1) > 0.02 or abs(test_wpeak2 - wpeak2) > 0.02:
        print(test_wpeak1 - wpeak1, test_wpeak2 - wpeak2)
        raise ValueError('Plasmon peak not correct ! ')

    if abs(test_Ipeak1 - Ipeak1) > 1 or abs(test_Ipeak2 - Ipeak2) > 1:
        print(Ipeak1 - test_Ipeak1, Ipeak2 - test_Ipeak2)
        raise ValueError('Please check spectrum strength ! ')



# %%
# teacher
import ase.visualize as viz
viz.view = lambda atoms, repeat=None: None

# %%
"""
# Curie temperature of CrI$_3$

In this exercise we will examine the magnetic poperties of 2D materials. In particular we would to determine the Curie temperature from first principles and try to predict a new 2D material that exhibits magnetic order above room temperature.

In 2017, ferromagnetic order was observed in a monolayer of CrI$_3$ below 45 K (*Nature* **546** 270 (2017)). It comprises the first demonstration of magnetic order in a 2D material and has received a lot of attention due to the peculiar properties of magnetism in 2D. The physics of magnetic order in 2D is rather different than in 3D and in order to understand what is going on we will need to introduce a bit of theory. But before we get to that let us get started with the calculations.
"""

# %%
"""
## DFT calculation - finding the atomic structure of CrI$_3$

We start by setting up the atomic structure of a CrI$_3$ monolayer and optimize the atomic positions and unit cell. There are two formula units in the minimal unit cell and the magnetic moments are large situated at the Cr atoms. A spin-polarized calculation is initiated by specifying the initial magnetic moments of the all the atoms in units of $\mu_B$.

1. What do you expect for the magnetic moment on th Cr atoms? (Hint: Use Hund's rule. The electronic configuration of a Cr atom is [Ar]3d$^5$4s$^1$ and each of the I atoms will still one electron)

Try to understand the individual lines in the input cell below and run it. The calculation will open the ase gui that shows the initial atomic structure. You may, for example, try to repeat the structure (under view) to get a better feeling for the material. Then look at the text output below the cell and answer the following questions:

1.  How many electrons are used in the calculation? Which valence states of Cr and I are included?
2.  What is the total magnetic moment after the first DFT calculation and on which atoms are the magnetic moments located?
3.  What is the number of irreducible k-points (k-points not related by symmetry) in the calculation?
4.  What is the maximum force on the atoms after the first DFT calculation? Does it become smaller after subsequent calculations?

In order to get the script running fast we have set a few of the parameters at values, which are expected to produce a somewhat inaccurate result. Can you identify the parameters that need to be converged/modifies in order to produce a more accurate ground state?

Leave the script running and continue with the theory section below.
"""

# %%
# %load summerschool.py
from gpaw import GPAW, PW
from ase import Atoms
from ase.visualize import view
from ase.optimize import BFGS
from ase.constraints import UnitCellFilter

a = 7.0
c = 10.0
S = 3 / 2  # student: S = ???
m = S * 2                              # Magnetic moment in Bohr magnetons
cell = [[a, 0, 0],
        [-0.5 * a, 3**0.5 / 2 * a, 0],
        [0, 0, c]]                     # Unit cell in \AA
scaled_pos = [[2 / 3., 1 / 3., 0.5],
              [1 / 3., 2 / 3., 0.5],
              [0.6, 1.0, 0.35],
              [0.4, 0.4, 0.35],
              [0.0, 0.6, 0.35],
              [0.4, 1.0, 0.65],
              [0.6, 0.6, 0.65],
              [0.0, 0.4, 0.65]]         # Positions in unit cell vectors
a = Atoms('Cr2I6', cell=cell, scaled_positions=scaled_pos, pbc=True)
a.set_initial_magnetic_moments([m, m, 0, 0, 0, 0, 0, 0])
view(a)

Nk = 2
calc = GPAW(mode=PW(200),
                xc='PBE',
                convergence={'density': 0.001, 'eigenstates': 0.1},
                kpts=(Nk, Nk, 1))
a.calc = calc

uf = UnitCellFilter(a, mask=[1, 1, 0, 0, 0, 1])
opt = BFGS(uf)
opt.run(fmax=0.2)
calc.write('CrI3_relaxed.gpw')

# %%
"""
## A bit of theory

### The Heisenberg model

If we want to calculate the Curie temperature of CrI$_3$ it is, in principle, clear what we have to do. We should calculate the magnetization as a function of temperature using standards methods from statistical physics, and record the temperature where the magnetization vanishes. Unfortunately, this requires knowledge of all the excited states of the system, which we do not have access to. In particular, the magnetization will be dominated by collective magnetic exciations, and these are not directly accessible from the Kohn-Sham spectrum produced by our DFT calculations.

Instead we will consider the Heisenberg Hamiltonian, which captures the basic physics of typical spin systems. It is given by

$$H = -\frac{1}{2}\sum_{ij}J_{ij}\mathbf{S}_i\cdot \mathbf{S}_j,$$

where $\mathbf{S}_i$ denotes the spin operator at site $i$ in units of $\hbar$ and $J_{ij}$ are magnetic exchange coupling constants. If we want to model a real material with the Heisenberg model we then need to identify a set of magnetic sites and calculate the exchange coupling constants $J_{ij}$.

1.   What are the magnetic sites of CrI$_3$?
2.   How many nearest neighbors does each magnetic site have?
3.   What are the possible values of $S_i^z$ for a magnetic site in CrI$_3$?
4.   What are the unit of $J_{ij}?$

In the following we will assume that the physics is dominated by neareast neighbor interactions such that $J_{ij}\equiv J$ if atoms $i$ and $j$ are nearest neighbors and zero otherwise. In 3D systems a reasonable estimate of the Curie temperature can be obtained from mean-field theory as

$$T_c^{\mathrm{MF}}=\frac{nS(S+1)J}{3k_B},$$

where $k_B$ is Boltzmann's constant, $n$ is the number of nearest neighbors, and $S$ is the maximum value of $S^z_i$.
"""

# %%
"""
## DFT calculation of $J$

We now want to make a first principles calculation of the nearest neighbor exchange coupling constant $J$. Since the exchange coupling essentially measures the energy difference between aligned and anti-aligned spin configurations we can obtain $J$ by considering the energy difference between a ferromagnetic and an anti-ferromagnetic calculation. Note that both can be obtained as DFT ground states subjected to different spin constraints. For the CrI$_3$ system, $J$ can calculated as

$$J=\frac{E_{\mathrm{AFM}}-E_{\mathrm{FM}}}{3S^2},$$

where $E_{\mathrm{FM}}$ is the energy of the anti-ferromagnetic configuration and $E_{\mathrm{AFM}}$ is the energy of the ferromagnetic configuration. Try to derive this expression from the Heisenberg model with classical spins. In particular, where did the factor of 3 come from?

We have compiled a database of various 2D materials at https://cmrdb.fysik.dtu.dk/?project=c2db, which are relaxed with the PBE functional. We will therefore refrain from doing a full coverged geometry optimization and simply download the optimized structure from the database. Search the database for CrI$_3$ and select the ferromagnetic structure (The present material has the prototype BiI$_3$). You can now look at various properties of the material like band structure and stability if you like. Download the .xyz filem, save it as CrI3.xyz, and run the cell below to obtain a .gpw file containing a converged ferromagnetic calculation. The calculation will take about 30 minutes. To speed it up, you can submit this and the next job to DTU computers on multiple CPU cores, i.e. parallelize over 8 k-points. Follow instructions [here](https://wiki.fysik.dtu.dk/gpaw/summerschools/summerschool18/submitting.html). If the relaxation in the cell above did not finish you may kill it. It is not crucial to complete the rest of the exercise. Continue with the theory below while you wait for the calculations to finish.
"""

# %%
from ase.io import read
from gpaw import GPAW, PW

a = read('CrI3.xyz')
a.set_initial_magnetic_moments([m, m, 0, 0, 0, 0, 0, 0])

Nk = 4
calc = GPAW(mode=PW(400), xc='PBE', kpts=(Nk, Nk, 1), symmetry='off')
a.calc = calc
a.get_potential_energy()
calc.write('CrI3_fm.gpw')


# %%
"""
Now run the cell again but change the initial magnetic configuration such that it becomes anti-ferromagnetic (remember to change the name of the .gpw file so you do not overwrite the CrI3_fm.gpw). In practise we do not need to constrain the anti-ferromagnetic calculation because it comprises a local minimum, but check the magnetic moments in the output to verify that we indeed ended up with an anti-ferromagnetic state!
"""

# %%
# teacher:
a = read('CrI3.xyz')
a.set_initial_magnetic_moments([m, -m, 0, 0, 0, 0, 0, 0])

Nk = 4
calc = GPAW(mode=PW(400), xc='PBE', kpts=(Nk, Nk, 1))
a.calc = calc
a.get_potential_energy()
calc.write('CrI3_afm.gpw')

# %%
"""
$J$ and $T_c^{\mathrm{MF}}$ can then be evaluated by
"""

# %%
calc_fm = GPAW('CrI3_fm.gpw', txt=None)      # Ferromagnetic calculation
calc_afm = GPAW('CrI3_afm.gpw', txt=None)    # Anti-ferromagnetic calculation

N = 3
E_fm = calc_fm.get_potential_energy() / 2    # Energy per site
E_afm = calc_afm.get_potential_energy() / 2  # Energy per site
dE = E_afm - E_fm

J = dE / S**2 / 3  # student: J = ???
print(f'J = {J * 1000:1.3f} meV')

from ase.units import kB                     # Boltzmann's constant in eV/K
T_c = N * J * S * (S + 1) / 3 / kB  # student: T_c = ???
print(f'T_c(MF) = {T_c:1.1f} K')

# %%
"""
## More theory

### The Mermin-Wagner theorem
You should obtain a value of $T_c$, which is on the order of 100 K. This is much larger than the experimental value.  However in 2D materials mean-field theory fails miserably and the result cannot be trusted. In fact, at finite temperatures the Heisenberg model stated above does not exhibit magnetic order in 2D. The reason is that entropy tends to dominate over internal energy in 2D, such that the free energy is always minimized by disordered configurations at finite temperatures. This is the content of the Mermin-Wagner theorem which states that:

*Continuous symmetries cannot be spontaneously broken at finite temperature in systems with short-range interactions in dimensions $d\le2$*.

The Heisenberg model above has a continuous rotational symmetry of the spins and magnetic order is obtained by choosing a certain direction for all the spins. This is known as a spontaneously broken symmetry, because the magnetically ordered state has broken the symmetry of the Hamiltonian. The direction of magnetization is arbitrary and can be rotated at no energy cost, but the spins are aligned with respect to each other and the system is therefore magnetically ordered.

In the Heisenberg model it is straightforward to calculate the magnetic excitations in the spin-wave approximation, which yields the dispersion

$$\varepsilon(q)=aq^2$$

for small values of $q$. The excitations are bosons and the magnetization can thus be calculated from

$$M=M_0 - \int_0^\infty\frac{g(\epsilon)d\varepsilon}{e^{\varepsilon/k_bT} - 1}$$

where $M_0$ is the ground state magnetization and $g(\varepsilon)$ is the density of states.

1.  Show that the integral diverges for $d\le2$ in the lower limit. (Hint you should start by calculating $g(\varepsilon)$. If you do not remember how to do that we give it here: $g(\varepsilon)=b\varepsilon^{(d-2)/2}$, where $d$ is the dimension of the problem).

The divergence of the integral in $d\le2$ signals that the ground state is unstable and comprises an example of the Mermin-Wagner theorem: in $d\le2$, the free energy is always dominated by entropy at finite temperatures and magnetic order cannot be maintained if a material has rotational symmetry of the spins. The instability in 2D is closely related to the vanishing energy of magnetic excitations in the limit of $q\rightarrow0$, which is a consequence of the rotational symmetry.

### Magnetic anisotropy
The consequence of the Mermin Wagner theorem is that magnetic order in 2D is only possible if the rotational symmetry of the spins is *explicitly broken*. That is, there must be an additional term in the Hamiltonian that breaks the symmetry. Such a term can be provided by spin-orbit coupling which couples the spin to the lattice through the electronic orbitals.
We assume that CrI$_3$ is isotropic in the plane of the monolayer and introduce an anisotropy term in the Heisenberg Hamiltonian of the form

$$H_{\mathrm{ani}}=A\sum_i(S_i^z)^2,$$

where we have chosen the $z$-direction to be orthogonal to the plane. Try to
describe the physics of this term in the cases of $A<0$ and $A>0$. Does the
term fully break rotational symmetry of the ground state in both cases?

## Magnetic anisotropy from DFT

In the cell below the magnetic anisotropy is calculated for the ferromagnetic
ground state. The function calculate_band_energy() will return the sum over
occupied SOC corrected states. This energy will depend on the direction of
the spins, which is specified by the polar and azimuthal angles $\theta$ and
$\varphi$ respectively.

What is the sign of $A$ in the Hamiltonian above? Does spin-orbit coupling
break the rotational symmetry of the ground state?

"""

# %%
from gpaw.spinorbit import soc_eigenstates

e_x = soc_eigenstates(calc_fm, theta=90, phi=0).calculate_band_energy() / 2
e_y = soc_eigenstates(calc_fm, theta=90, phi=90).calculate_band_energy() / 2
e_z = soc_eigenstates(calc_fm, theta=0, phi=0).calculate_band_energy() / 2
de_zx = e_z - e_x
de_zy = e_z - e_y
print(f'dE_zx = {de_zx * 1000:1.3f} meV')
print(f'dE_zy = {de_zy * 1000:1.3f} meV')
A = de_zx  # student: A = ???


# %%
"""
We can also plot the total energy of the ground state as a function of angle
with the out of plane direction. This is done by the cell below.

Run the cell and inspect the plot. Does it look like you would expect?
"""

# %%
import matplotlib.pyplot as plt
import numpy as np

thetas = np.linspace(0, 180, 13)

e_n = []
for theta in thetas:
    soc = soc_eigenstates(calc_fm, theta=theta, phi=0)
    e_n.append(soc.calculate_band_energy() / 2)

e_n = np.array(e_n) - e_n[0]

plt.figure()
plt.plot(thetas, e_n * 1000, 'o-')
plt.xlabel(r'$\theta$', size=18)
plt.ylabel('E [meV]', size=18)


# %%
"""
Now we have obtained the the anisotropy constant A. But how do we get the critical temperature if we cannot apply mean-field theory? One
way is to perform Monte-Carlo simulations of the Heisenberg model at different temperature and find the point where the total magnetization
vanishes. The results are well approximated by the expression

$$T_c=T_c^{\mathrm{Ising}}\tanh^{1/4}\Big[\frac{6}{N}\log\Big(1-0.033\frac{A}{J}\Big)\Big],$$

where $N$ is the number of nearest neighbors and and $T_c^{\mathrm{Ising}}=1.52\cdot S^2J/k_B$ is the critical temperature of the
corresponding Ising model. Calculate the critical temperature from this expression using the values of $A$ and $J$ found above. Do with
python in the cell below.

"""

# %%
from numpy import tanh, log

T0 = 1.52
T_c = T0 * S**2 * J / kB * (tanh(2 * log(1 - 0.033 * A / J)))**(0.25)  # student: T_c = ???
print(f'T_c = {T_c:.1f} K')


# %%
"""
The result for $T_c$ should be in reasonable agreement with the experimental value. Of course, one should carefully check the convergence
of all calculations in the present notebook. In fact a converged calculation yields $J = 3.1$ meV and $A=-0.38$ meV, which results in $T_c
= 36$ K.
"""



import numpy as np

import gpaw.mpi as mpi
from gpaw.response.susceptibility import FourComponentSusceptibilityTensor

FCST = FourComponentSusceptibilityTensor


class TransverseMagneticSusceptibility(FCST):
    """Class calculating the transverse magnetic susceptibility
    and related physical quantities."""

    def __init__(self, *args, **kwargs):
        assert kwargs['fxc'] == 'ALDA'

        # Enable scaling to fit to Goldstone theorem
        if 'fxckwargs' in kwargs and 'fxc_scaling' in kwargs['fxckwargs']:
            self.fxc_scaling = kwargs['fxckwargs']['fxc_scaling']
        else:
            self.fxc_scaling = None

        FCST.__init__(self, *args, **kwargs)

    def get_macroscopic_component(self, spincomponent, q_c, frequencies,
                                  filename=None, txt=None):
        """Calculates the spatially averaged (macroscopic) component of the
        transverse magnetic susceptibility and writes it to a file.
        
        Parameters
        ----------
        spincomponent : str
            '+-': calculate chi+-, '-+: calculate chi-+
        q_c, frequencies, filename, txt : see gpaw.response.susceptibility

        Returns
        -------
        see gpaw.response.susceptibility
        """
        assert spincomponent in ['+-', '-+']

        return FCST.get_macroscopic_component(self, spincomponent, q_c,
                                              frequencies, filename=filename,
                                              txt=txt)

    def get_component_array(self, spincomponent, q_c, frequencies,
                            array_ecut=50, filename=None, txt=None):
        """Calculates a specific spin component of the
        transverse magnetic susceptibility and writes it to a file.
        
        Parameters
        ----------
        spincomponent : str
            '+-': calculate chi+-, '-+: calculate chi-+
        q_c, frequencies,
        array_ecut, filename, txt : see gpaw.response.susceptibility

        Returns
        -------
        see gpaw.response.susceptibility
        """
        assert spincomponent in ['+-', '-+']

        return FCST.get_component_array(self, spincomponent, q_c,
                                        frequencies, array_ecut=array_ecut,
                                        filename=filename, txt=txt)

    def _calculate_component(self, spincomponent, pd, wd):
        """Calculate a transverse magnetic susceptibility element.

        Returns
        -------
        pd, wd, chiks_wGG, chi_wGG : see gpaw.response.susceptibility
        """
        chiks_wGG = self.calculate_ks_component(spincomponent, pd,
                                                wd, txt=self.cfd)
        Kxc_GG = self.get_xc_kernel(spincomponent, pd,
                                    chiks_wGG=chiks_wGG, txt=self.cfd)

        chi_wGG = self.invert_dyson(chiks_wGG, Kxc_GG)

        return pd, wd, chiks_wGG, chi_wGG

    def get_xc_kernel(self, spincomponent, pd, chiks_wGG=None, txt=None):
        """Get the exchange correlation kernel."""
        Kxc_GG = self.fxc(spincomponent, pd, txt=self.cfd)

        fxc_scaling = self.fxc_scaling

        if fxc_scaling is not None:
            assert isinstance(fxc_scaling[0], bool)
            if fxc_scaling[0]:
                if fxc_scaling[1] is None:
                    assert pd.kd.gamma
                    print('Finding rescaling of kernel to fulfill the '
                          'Goldstone theorem', file=self.fd)
                    fxc_scaling[1] = find_goldstone_scaling(self.chiks.omega_w,
                                                            chiks_wGG, Kxc_GG,
                                                            world=self.world)

                assert isinstance(fxc_scaling[1], float)
                Kxc_GG *= fxc_scaling[1]

        self.fxc_scaling = fxc_scaling

        return Kxc_GG


def find_goldstone_scaling(omega_w, chi0_wGG, Kxc_GG, world=mpi.world):
    """Find a scaling of the kernel to move the magnon peak to omega=0."""
    wgs = np.abs(omega_w).argmin()
    if not np.allclose(omega_w[wgs], 0., rtol=1.e-8):
        raise ValueError("Frequency grid needs to include"
                         + " omega=0. to allow Goldstone scaling")

    fxcs = 1.

    # Only one rank, rgs, has omega=0 and finds rescaling
    nw = len(omega_w)
    mynw = (nw + world.size - 1) // world.size
    rgs, mywgs = wgs // mynw, wgs % mynw
    fxcsbuf = np.empty(1, dtype=float)
    if world.rank == rgs:
        chi0_GG = chi0_wGG[mywgs]
        chi_GG = np.dot(np.linalg.inv(np.eye(len(chi0_GG)) +
                                      np.dot(chi0_GG, Kxc_GG * fxcs)),
                        chi0_GG)
        # Scale so that kappaM=0 in the static limit (omega=0)
        kappaM = (chi0_GG[0, 0] / chi_GG[0, 0]).real
        # If kappaM > 0, increase scaling (recall: kappaM ~ 1 - Kxc Re{chi_0})
        scaling_incr = 0.1 * np.sign(kappaM)
        while abs(kappaM) > 1.e-7 and abs(scaling_incr) > 1.e-7:
            fxcs += scaling_incr
            if fxcs <= 0.0 or fxcs >= 10.:
                raise Exception('Found an invalid fxc_scaling of %.4f' % fxcs)

            chi_GG = np.dot(np.linalg.inv(np.eye(len(chi0_GG)) +
                                          np.dot(chi0_GG, Kxc_GG * fxcs)),
                            chi0_GG)
            kappaM = (chi0_GG[0, 0] / chi_GG[0, 0]).real

            # If kappaM changes sign, change sign and refine increment
            if np.sign(kappaM) != np.sign(scaling_incr):
                scaling_incr *= -0.2
        fxcsbuf[:] = fxcs

    # Broadcast found rescaling
    world.broadcast(fxcsbuf, rgs)
    fxcs = fxcsbuf[0]

    return fxcs



import pickle
import numpy as np
from gpaw import GPAW, setup_paths
from gpaw.vdw import VDWFunctional

setup_paths.insert(0, '.')

d = np.linspace(3.0, 5.5, 11)
for symbol in ['Ar', 'Kr']:
    vdw = VDWFunctional(verbose=1)
    e = np.empty(11)
    de = np.empty(11)
    for i, r in enumerate(d):
        calc = GPAW('%s-dimer-%.2f.gpw' % (symbol, r), txt=None)
        e[i] = calc.get_atoms().get_potential_energy()
        de[i] = calc.get_xc_difference(vdw)
        print(i, e, de)
    calc = GPAW('%s-atom.gpw' % symbol, txt=None)
    e0 = calc.get_atoms().get_potential_energy()
    de0 = calc.get_xc_difference(vdw)
    print(e, de, e0, de0)
    pickle.dump((e, de, e0, de0), open(symbol + '.new.pckl', 'wb'))

e = np.empty(11)
de = np.empty(11)
vdw = VDWFunctional(verbose=1)
for i, r in enumerate(d):
    calc = GPAW('benzene-dimer-%.2f.gpw' % r, txt=None)
    e[i] = calc.get_atoms().get_potential_energy()
    de[i] = calc.get_xc_difference(vdw)
calc = GPAW('benzene.gpw', txt=None)
e0 = calc.get_atoms().get_potential_energy()
de0 = calc.get_xc_difference(vdw)
print(e, de, e0, de0)
pickle.dump((e, de, e0, de0), open('benzene.new.pckl', 'wb'))



from gpaw.response.g0w0 import G0W0

ecut = 40
gw = G0W0(calc='MoS2_fulldiag.gpw',
          xc='rALDA',
          av_scheme='wavevector',
          fxc_mode='GWG',
          bands=(8, 18),
          ecut=ecut,
          truncation='2D',
          nblocksmax=True,
          q0_correction=True,
          filename=f'MoS2_g0w0g_{ecut}',
          savepckl=True)

gw.calculate()




import os
import time
import numpy as np

from ase import Atoms
from ase.parallel import paropen
from ase.units import Hartree, Bohr
from ase.io import Trajectory
from ase.calculators.singlepoint import SinglePointCalculator
from gpaw import GPAW
from gpaw.mpi import world
from gpaw.tddft import TDDFT
from gpaw.tddft.ehrenfest import EhrenfestVelocityVerlet


name = 'na2_osc'

# Equilibrium distance in Ang cf. setups page for Na dimer
d_bond = 3.29  # ~3.0 during oscillation
d_disp = 0.1

# Timestep and expected oscillatory period in attoseconds
timestep = 5.0
period = 2.1e5  # ~19.7 meV cf. CRC Handbook of Phys. & Chem. #09_08_91

ndiv = int(np.ceil(0.1e3 / timestep))  # update stats every 0.1 fs
niter = ndiv * int(np.ceil(period / (ndiv * timestep)))

if __name__ == '__main__':
    if not os.path.isfile(name + '_gs.gpw'):
        atoms = Atoms('Na2', positions=[(0, 0, 0), (0, 0, d_bond + d_disp)])
        atoms.set_pbc(False)
        atoms.center(vacuum=6.0)
        cell_c = np.sum(atoms.get_cell()**2, axis=1)**0.5
        N_c = 16 * np.round(cell_c / (0.25 * 16))
        calc = GPAW(gpts=N_c, nbands=1, basis='dzp', setups={'Na': '1'},
                    txt=name + '_gs.txt')
        atoms.calc = calc
        atoms.get_potential_energy()
        calc.write(name + '_gs.gpw', mode='all')
        del atoms, calc
        time.sleep(10)

    while not os.path.isfile(name + '_gs.gpw'):
        print('Node %d waiting for file...' % world.rank)
        time.sleep(10)
    world.barrier()

    tdcalc = TDDFT(name + '_gs.gpw', txt=name + '_td.txt', propagator='EFSICN')
    ehrenfest = EhrenfestVelocityVerlet(tdcalc)
    traj = Trajectory(name + '_td.traj', 'w', tdcalc.get_atoms())

    t0 = time.time()
    f = paropen(name + '_td.log', 'w')
    for i in range(1, niter + 1):
        ehrenfest.propagate(timestep)

        if i % ndiv == 0:
            rate = 60 * ndiv / (time.time() - t0)
            ekin = tdcalc.atoms.get_kinetic_energy()
            epot = tdcalc.get_td_energy() * Hartree
            F_av = ehrenfest.F * Hartree / Bohr
            s = 'i=%06d (%6.2f min^-1), ekin=%13.9f, epot=%13.9f, etot=%13.9f'
            print(s % (i, rate, ekin, epot, ekin + epot), file=f)
            t0 = time.time()

            # Hack to prevent calls to GPAW::get_potential_energy when saving
            spa = tdcalc.get_atoms()
            spc = SinglePointCalculator(spa, energy=epot, forces=F_av)
            spa.calc = spc
            traj.write(spa)
    f.close()
    traj.close()



from myqueue.task import task


def create_tasks():
    return [
        task('dipole.py@4:1h'),
        task('pwdipole.py@4:5m'),
        task('plot.py', deps='dipole.py,pwdipole.py'),
        task('check.py', deps='dipole.py,pwdipole.py')]



import pytest
from gpaw.basis_data import Basis
from gpaw.mpi import world
from gpaw.lcao.generate_ngto_augmented import do_nao_ngto_basis

# Generate file with GTO parameters
pytestmark = pytest.mark.skipif(world.size > 1,
                                reason='world.size > 1')


def test_lcao_generate_ngto(in_tmp_dir):
    with open('gbs.txt', 'w') as f:
        def w(s):
            f.write('%s\n' % s)
        w('****')
        w('H     0')
        w('S   1   1.00')
        w('      0.1000000              1.0000000')
        w('P   1   1.00')
        w('      0.0500000              1.0000000')
        w('D   1   1.00')
        w('      0.5000000              1.0000000')
        w('****')

    # Run the generator script in order to keep the syntax up-to-date
    do_nao_ngto_basis('H', 'LDA', 'sz', 'gbs.txt', 'NAO+NGTO')

    # Check that the generated basis contains the correct number of functions
    basis = Basis('H', 'NAO+NGTO', readxml=False)
    basis.read_xml('H.NAO+NGTO.sz.basis')
    bf_j = basis.bf_j

    assert len(bf_j) == 1 + 3

    with open('gbs.txt', 'w') as f:
        def w(s):
            f.write('%s\n' % s)
        w('****')
        w('C     0')
        w('S   1   1.00')
        w('      1.596000D-01           1.000000D+00')
        w('S   1   1.00')
        w('      0.0469000              1.0000000')
        w('P   4   1.00')
        w('      9.439000D+00           3.810900D-02')
        w('      2.002000D+00           2.094800D-01')
        w('      5.456000D-01           5.085570D-01')
        w('      1.517000D-01           4.688420D-01')
        w('P   1   1.00')
        w('      1.517000D-01           1.000000D+00')
        w('P   1   1.00')
        w('      0.0404100              1.0000000')
        w('D   1   1.00')
        w('      5.500000D-01           1.0000000')
        w('D   1   1.00')
        w('      0.1510000              1.0000000')
        w('****')

    do_nao_ngto_basis('C', 'LDA', 'dzp', 'gbs.txt', 'DZP+NGTO')

    basis = Basis('C', 'DZP+NGTO', readxml=False)
    basis.read_xml('C.DZP+NGTO.dzp.basis')
    bf_j = basis.bf_j

    assert len(bf_j) == 5 + 7
